########
#
# run_tiled_inference.py
#
# Run inference on a folder, fist splitting each image up into tiles of size
# MxN (typically the native inference size of your detector), writing those
# tiles out to a temporary folder, then de-duplicating the results before merging
# them back into a set of detections that make sense on the original images.
#
# This approach will likely fail to detect very large animals, so if you expect both large 
# and small animals (in terms of pixel size), this script is best used in 
# conjunction with a traditional inference pass that looks at whole images.
#
# Currently requires temporary storage at least as large as the input data, generally
# a lot more than that (depending on the overlap between adjacent tiles).  This is 
# inefficient, but easy to debug.
#
# Programmatic invocation supports using YOLOv5's inference scripts (and test-time
# augmentation); the command-line interface only supports standard inference right now.
#
########

#%% Imports and constants

import os
import json

from tqdm import tqdm

from detection.run_inference_with_yolov5_val import YoloInferenceOptions,run_inference_with_yolo_val
from detection.run_detector_batch import load_and_run_detector_batch,write_results_to_file

import torch
from torchvision import ops

from md_utils import path_utils
from md_visualization import visualization_utils as vis_utils

default_patch_overlap = 0.5
patch_jpeg_quality = 95

# This isn't NMS in the usual sense of redundant model predictions; this is being
# used to de-duplicate predictions from overlapping patches.
nms_iou_threshold = 0.45

default_tile_size = [1280,1280]

default_n_patch_extraction_workers = 1
parallelization_uses_threads = False


#%% Support functions

def get_patch_boundaries(image_size,patch_size,patch_stride=None):
    """
    Get a list of patch starting coordinates (x,y) given an image size (w,h)
    and a stride (x,y).  Stride defaults to half the patch size.
    
    patch_stride can also be a single float, in which case that is interpreted 
    as the stride relative to the patch size (0.1 == 10% stride).
    
    Patch size is guaranteed, stride may deviate to make sure all pixels are covered.
    I.e., we move by regular strides until the current patch walks off the right/bottom,
    at which point it backs up to one patch from the end.  So if your image is 15
    pixels wide and you have a stride of 10 pixels, you will get starting positions 
    of 0 (from 0 to 9) and 5 (from 5 to 14).
    """
    
    if patch_stride is None:
        patch_stride = (round(patch_size[0]*(1.0-default_patch_overlap)),
                        round(patch_size[1]*(1.0-default_patch_overlap)))
    elif isinstance(patch_stride,float):
        patch_stride = (round(patch_size[0]*(patch_stride)),
                        round(patch_size[1]*(patch_stride)))
        
    image_width = image_size[0]
    image_height = image_size[1]
    
    assert patch_size[0] <= image_size[0], 'Patch width {} is larger than image width {}'.format(
        patch_size[0],image_size[0])
    assert patch_size[1] <= image_size[1], 'Patch height {} is larger than image height {}'.format(
        patch_size[1],image_size[1])
    
    def add_patch_row(patch_start_positions,y_start):
        """
        Add one row to our list of patch start positions, i.e.
        loop over all columns.
        """
        
        x_start = 0; x_end = x_start + patch_size[0] - 1
        
        while(True):
            
            patch_start_positions.append([x_start,y_start])
            
            # If this patch put us right at the end of the last column, we're done
            if x_end == image_width - 1:
                break
            
            # Move one patch to the right
            x_start += patch_stride[0]
            x_end = x_start + patch_size[0] - 1
             
            # If this patch flows over the edge, add one more patch to cover
            # the pixels on the end, then we're done.
            if x_end > (image_width - 1):
                overshoot = (x_end - image_width) + 1
                x_start -= overshoot
                x_end = x_start + patch_size[0] - 1
                patch_start_positions.append([x_start,y_start])
                break
        
        # ...for each column
        
        return patch_start_positions
        
    patch_start_positions = []
    
    y_start = 0; y_end = y_start + patch_size[1] - 1
    
    while(True):
    
        patch_start_positions = add_patch_row(patch_start_positions,y_start)
        
        # If this patch put us right at the bottom of the lats row, we're done
        if y_end == image_height - 1:
            break
        
        # Move one patch down
        y_start += patch_stride[1]
        y_end = y_start + patch_size[1] - 1
        
        # If this patch flows over the bottom, add one more patch to cover
        # the pixels at the bottom, then we're done
        if y_end > (image_height - 1):
            overshoot = (y_end - image_height) + 1
            y_start -= overshoot
            y_end = y_start + patch_size[1] - 1
            patch_start_positions = add_patch_row(patch_start_positions,y_start)
            break
    
    # ...for each row
    
    for p in patch_start_positions:
        assert p[0] >= 0 and p[1] >= 0 and p[0] <= image_width and p[1] <= image_height, \
        'Patch generation error (illegal patch {})'.format(p)
        
    # The last patch should always end at the bottom-right of the image
    assert patch_start_positions[-1][0]+patch_size[0] == image_width, \
        'Patch generation error (last patch does not end on the right)'
    assert patch_start_positions[-1][1]+patch_size[1] == image_height, \
        'Patch generation error (last patch does not end at the bottom)'
    
    # All patches should be unique
    patch_start_positions_tuples = [tuple(x) for x in patch_start_positions]
    assert len(patch_start_positions_tuples) == len(set(patch_start_positions_tuples)), \
        'Patch generation error (duplicate start position)'
    
    return patch_start_positions

# ...get_patch_boundaries()


def patch_info_to_patch_name(image_name,patch_x_min,patch_y_min):
    
    patch_name = image_name + '_' + \
        str(patch_x_min).zfill(4) + '_' + str(patch_y_min).zfill(4)
    return patch_name


def extract_patch_from_image(im,patch_xy,patch_size,
                             patch_image_fn=None,patch_folder=None,image_name=None,overwrite=True):
    """
    Extracts a patch from the provided image, writing the patch out to patch_image_fn.
    [im] can be a string or a PIL image.
    
    patch_xy is a length-2 tuple specifying the upper-left corner of the patch.
    
    image_name and patch_folder are only required if patch_image_fn is None.
    
    Returns a dictionary with fields xmin,xmax,ymin,ymax,patch_fn.
    """
    
    if isinstance(im,str):
        pil_im = vis_utils.open_image(im)
    else:
        pil_im = im
        
    patch_x_min = patch_xy[0]
    patch_y_min = patch_xy[1]
    patch_x_max = patch_x_min + patch_size[0] - 1
    patch_y_max = patch_y_min + patch_size[1] - 1

    # PIL represents coordinates in a way that is very hard for me to get my head
    # around, such that even though the "right" and "bottom" arguments to the crop()
    # function are inclusive... well, they're not really.
    #
    # https://pillow.readthedocs.io/en/stable/handbook/concepts.html#coordinate-system
    #
    # So we add 1 to the max values.
    patch_im = pil_im.crop((patch_x_min,patch_y_min,patch_x_max+1,patch_y_max+1))
    assert patch_im.size[0] == patch_size[0]
    assert patch_im.size[1] == patch_size[1]

    if patch_image_fn is None:
        assert patch_folder is not None,\
            "If you don't supply a patch filename to extract_patch_from_image, you need to supply a folder name"
        patch_name = patch_info_to_patch_name(image_name,patch_x_min,patch_y_min)
        patch_image_fn = os.path.join(patch_folder,patch_name + '.jpg')
    
    if os.path.isfile(patch_image_fn) and (not overwrite):
        pass
    else:        
        patch_im.save(patch_image_fn,quality=patch_jpeg_quality)
    
    patch_info = {}
    patch_info['xmin'] = patch_x_min
    patch_info['xmax'] = patch_x_max
    patch_info['ymin'] = patch_y_min
    patch_info['ymax'] = patch_y_max
    patch_info['patch_fn'] = patch_image_fn
    
    return patch_info


def in_place_nms(md_results, iou_thres=0.45, verbose=True):
    """
    Run torch.ops.nms in-place on MD-formatted detection results    
    """
    
    n_detections_before = 0
    n_detections_after = 0
    
    # i_image = 18; im = md_results['images'][i_image]
    for i_image,im in tqdm(enumerate(md_results['images']),total=len(md_results['images'])):
        
        if len(im['detections']) == 0:
            continue
    
        boxes = []
        scores = []
        
        n_detections_before += len(im['detections'])
        
        # det = im['detections'][0]
        for det in im['detections']:
            
            # Using x1/x2 notation rather than x0/x1 notation to be consistent
            # with the Torch documentation.
            x1 = det['bbox'][0]
            y1 = det['bbox'][1]
            x2 = det['bbox'][0] + det['bbox'][2]
            y2 = det['bbox'][1] + det['bbox'][3]
            box = [x1,y1,x2,y2]
            boxes.append(box)
            scores.append(det['conf'])

        # ...for each detection
        
        t_boxes = torch.tensor(boxes)
        t_scores = torch.tensor(scores)
        
        box_indices = ops.nms(t_boxes,t_scores,iou_thres).tolist()
        
        post_nms_detections = [im['detections'][x] for x in box_indices]
        
        assert len(post_nms_detections) <= len(im['detections'])
        
        im['detections'] = post_nms_detections
        
        n_detections_after += len(im['detections'])
        
    # ...for each image
    
    if verbose:
        print('NMS removed {} of {} detections'.format(
            n_detections_before-n_detections_after,
            n_detections_before))
        
# ...in_place_nms()


def _extract_tiles_for_image(fn_relative,image_folder,tiling_folder,patch_size,patch_stride,overwrite):
    """
    Extract tiles for a single image
    
    Not really a standalone function; isolated from the main function to simplify
    multiprocessing.
    """
    
    fn_abs = os.path.join(image_folder,fn_relative)
    
    image_name = path_utils.clean_filename(fn_relative,char_limit=None,force_lower=True)
    
    # Open the image
    im = vis_utils.open_image(fn_abs)
    image_size = [im.width,im.height]
            
    # Generate patch boundaries (a list of [x,y] starting points)
    patch_boundaries = get_patch_boundaries(image_size,patch_size,patch_stride)        
    
    # Extract patches
    #
    # patch_xy = patch_boundaries[0]
    patches = []
    
    for patch_xy in patch_boundaries:
        
        patch_info = extract_patch_from_image(im,patch_xy,patch_size,
                                 patch_folder=tiling_folder,
                                 image_name=image_name,
                                 overwrite=overwrite)
        patch_info['source_fn'] = fn_relative
        patches.append(patch_info)
        
    image_patch_info = {}
    image_patch_info['patches'] = patches
    image_patch_info['image_fn'] = fn_relative
    
    return image_patch_info
    
    
#%% Main function
    
def run_tiled_inference(model_file, image_folder, tiling_folder, output_file,
                        tile_size_x=1280, tile_size_y=1280, tile_overlap=0.5,
                        checkpoint_path=None, checkpoint_frequency=-1, remove_tiles=False, 
                        yolo_inference_options=None,
                        n_patch_extraction_workers=default_n_patch_extraction_workers,
                        overwrite_tiles=True):
    """
    Run inference using [model_file] on the images in [image_folder], fist splitting each image up 
    into tiles of size [tile_size_x] x [tile_size_y], writing those tiles to [tiling_folder],
    then de-duplicating the results before merging them back into a set of detections that make 
    sense on the original images and writing those results to [output_file].  
    
    [tiling_folder] can be any folder, but this function reserves the right to do whatever it wants
    within that folder, including deleting everything, so it's best if it's a new folder.  
    Conceptually this folder is temporary, it's just helpful in this case to not actually
    use the system temp folder, because the tile cache may be very large, 
    
    tile_overlap is the fraction of overlap between tiles.
    
    Optionally removes the temporary tiles.
    
    if yolo_inference_options is supplied, it should be an instance of YoloInferenceOptions; in 
    this case the model will be run with run_inference_with_yolov5_val.  This is typically used to 
    run the model with test-time augmentation.          
    """    

    ##%% Validate arguments
    
    assert tile_overlap < 1 and tile_overlap >= 0, \
        'Illegal tile overlap value {}'.format(tile_overlap)
    
    patch_size = [tile_size_x,tile_size_y]
    patch_stride = (round(patch_size[0]*(1.0-tile_overlap)),
                    round(patch_size[1]*(1.0-tile_overlap)))
    
    os.makedirs(tiling_folder,exist_ok=True)
    
    
    ##%% List files
    
    image_files_relative = path_utils.find_images(image_folder, recursive=True, return_relative_paths=True)    
    assert len(image_files_relative) > 0, 'No images found in folder {}'.format(image_folder)
    
    
    ##%% Generate tiles
    
    all_image_patch_info = None
    
    print('Extracting patches from {} images'.format(len(image_files_relative)))
    
    n_workers = n_patch_extraction_workers
    
    if n_workers <= 1:
        
        all_image_patch_info = []
        
        # fn_relative = image_files_relative[0]        
        for fn_relative in tqdm(image_files_relative):        
            image_patch_info = \
                _extract_tiles_for_image(fn_relative,image_folder,tiling_folder,patch_size,patch_stride,
                                         overwrite=overwrite_tiles)
            all_image_patch_info.append(image_patch_info)
            
    else:
        
        from multiprocessing.pool import ThreadPool
        from multiprocessing.pool import Pool
        from functools import partial

        if n_workers > len(image_files_relative):
            
            print('Pool of {} requested, but only {} images available, reducing pool to {}'.\
                  format(n_workers,len(image_files_relative),len(image_files_relative)))
            n_workers = len(image_files_relative)
                                
        if parallelization_uses_threads:
            pool = ThreadPool(n_workers); poolstring = 'threads'                
        else:
            pool = Pool(n_workers); poolstring = 'processes'

        print('Starting patch extraction pool with {} {}'.format(n_workers,poolstring))
        
        all_image_patch_info = list(tqdm(pool.imap(
                partial(_extract_tiles_for_image,
                        image_folder=image_folder,
                        tiling_folder=tiling_folder,
                        patch_size=patch_size,
                        patch_stride=patch_stride,
                        overwrite=overwrite_tiles), 
                image_files_relative),total=len(image_files_relative)))
        
    # ...for each image
        
    # Write tile information to file; this is just a debugging convenience
    folder_name = path_utils.clean_filename(image_folder,force_lower=True)
    if folder_name.startswith('_'):
        folder_name = folder_name[1:]
        
    tile_cache_file = os.path.join(tiling_folder,folder_name + '_patch_info.json')
    with open(tile_cache_file,'w') as f:
        json.dump(all_image_patch_info,f,indent=1)
    
    
    ##%% Run inference on tiles
    
    if yolo_inference_options is not None:
        
        patch_level_output_file = os.path.join(tiling_folder,folder_name + '_patch_level_results.json')
        
        if yolo_inference_options.model_filename is None:
            yolo_inference_options.model_filename = model_file
        else:
            assert yolo_inference_options.model_filename == model_file, \
                'Model file between yolo inference file ({}) and model file parameter ({})'.format(
                    yolo_inference_options.model_filename,model_file)
        
        yolo_inference_options.input_folder = tiling_folder
        yolo_inference_options.output_file = patch_level_output_file
        
        run_inference_with_yolo_val(yolo_inference_options)
        with open(patch_level_output_file,'r') as f:
            patch_level_results = json.load(f)
                    
    else:
        
        patch_file_names = []
        for im in all_image_patch_info:
            for patch in im['patches']:
                patch_file_names.append(patch['patch_fn'])
                
        inference_results = load_and_run_detector_batch(model_file, 
                                                        patch_file_names, 
                                                        checkpoint_path=checkpoint_path,
                                                        checkpoint_frequency=checkpoint_frequency,
                                                        quiet=True)
        
        patch_level_output_file = os.path.join(tiling_folder,folder_name + '_patch_level_results.json')
        
        patch_level_results = write_results_to_file(inference_results, 
                                                    patch_level_output_file, 
                                                    relative_path_base=tiling_folder, 
                                                    detector_file=model_file)
        
    
    ##%% Map patch-level detections back to the original images    
    
    # Map relative paths for patches to detections
    patch_fn_relative_to_results = {}
    for im in tqdm(patch_level_results['images']):
        patch_fn_relative_to_results[im['file']] = im

    image_level_results = {}
    image_level_results['info'] = patch_level_results['info']
    image_level_results['detection_categories'] = patch_level_results['detection_categories']
    image_level_results['images'] = []
    
    image_fn_relative_to_patch_info = { x['image_fn']:x for x in all_image_patch_info }
    
    # i_image = 0; image_fn_relative = image_files_relative[i_image]
    for i_image,image_fn_relative in tqdm(enumerate(image_files_relative),total=len(image_files_relative)):
        
        image_fn_abs = os.path.join(image_folder,image_fn_relative)
        assert os.path.isfile(image_fn_abs)
                
        output_im = {}
        output_im['file'] = image_fn_relative
        output_im['detections'] = []
            
        pil_im = vis_utils.open_image(image_fn_abs)        
        image_w = pil_im.size[0]
        image_h = pil_im.size[1]
        
        image_patch_info = image_fn_relative_to_patch_info[image_fn_relative]
        assert image_patch_info['patches'][0]['source_fn'] == image_fn_relative
        
        # Patches for this image
        patch_fn_abs_to_patch_info_this_image = {}
        
        for patch_info in image_patch_info['patches']:
            patch_fn_abs_to_patch_info_this_image[patch_info['patch_fn']] = patch_info
                
        # For each patch
        #
        # i_patch = 0; patch_fn_abs = list(patch_fn_abs_to_patch_info_this_image.keys())[i_patch]
        for i_patch,patch_fn_abs in enumerate(patch_fn_abs_to_patch_info_this_image.keys()):
            
            patch_fn_relative = os.path.relpath(patch_fn_abs,tiling_folder)
            patch_results = patch_fn_relative_to_results[patch_fn_relative]
            patch_info = patch_fn_abs_to_patch_info_this_image[patch_fn_abs]
            
            # patch_results['file'] is a relative path, and a subset of patch_info['patch_fn']
            assert patch_results['file'] in patch_info['patch_fn']
            
            patch_w = (patch_info['xmax'] - patch_info['xmin']) + 1
            patch_h = (patch_info['ymax'] - patch_info['ymin']) + 1
            assert patch_w == patch_size[0]
            assert patch_h == patch_size[1]
            
            # det = patch_results['detections'][0]
            for det in patch_results['detections']:
            
                bbox_patch_relative = det['bbox']
                xmin_patch_relative = bbox_patch_relative[0]
                ymin_patch_relative = bbox_patch_relative[1]
                w_patch_relative = bbox_patch_relative[2]
                h_patch_relative = bbox_patch_relative[3]
                
                # Convert from patch-relative normalized values to image-relative absolute values
                w_pixels = w_patch_relative * patch_w
                h_pixels = h_patch_relative * patch_h
                xmin_patch_pixels = xmin_patch_relative * patch_w
                ymin_patch_pixels = ymin_patch_relative * patch_h
                xmin_image_pixels = patch_info['xmin'] + xmin_patch_pixels
                ymin_image_pixels = patch_info['ymin'] + ymin_patch_pixels
                
                # ...and now to image-relative normalized values
                w_image_normalized = w_pixels / image_w
                h_image_normalized = h_pixels / image_h
                xmin_image_normalized = xmin_image_pixels / image_w
                ymin_image_normalized = ymin_image_pixels / image_h
                
                bbox_image_normalized = [xmin_image_normalized,
                                         ymin_image_normalized,
                                         w_image_normalized,
                                         h_image_normalized]
                
                output_det = {}
                output_det['bbox'] = bbox_image_normalized
                output_det['conf'] = det['conf']
                output_det['category'] = det['category']
                
                output_im['detections'].append(output_det)
                
            # ...for each detection
            
        # ...for each patch

        image_level_results['images'].append(output_im)
        
    # ...for each image    

    image_level_results_file_pre_nms = \
        os.path.join(tiling_folder,folder_name + '_image_level_results_pre_nms.json')
    with open(image_level_results_file_pre_nms,'w') as f:
        json.dump(image_level_results,f,indent=1)
        

    ##%% Run NMS
    
    in_place_nms(image_level_results,iou_thres=nms_iou_threshold)

    
    ##%% Write output file
    
    print('Saving image-level results (after NMS) to {}'.format(output_file))
    
    with open(output_file,'w') as f:
        json.dump(image_level_results,f,indent=1)

        
    ##%% Possibly remove tiles
        
    if remove_tiles:
        
        patch_file_names = []
        for im in all_image_patch_info:
            for patch in im['patches']:
                patch_file_names.append(patch['patch_fn'])
                
        for patch_fn_abs in patch_file_names:
            os.remove(patch_fn_abs)
        
    
    ##%% Return
    
    return image_level_results


#%% Interactive driver

if False:
    
    pass

    #%% Run tiled inference (in Python)
    
    model_file = os.path.expanduser('~/models/camera_traps/megadetector/md_v5.0.0/md_v5a.0.0.pt')
    image_folder = os.path.expanduser('~/data/KRU-test')
    tiling_folder = os.path.expanduser('~/tmp/tiling-test')
    output_file = os.path.expanduser('~/tmp/KRU-test-tiled.json')

    tile_size_x = 3000
    tile_size_y = 3000
    tile_overlap = 0.5
    checkpoint_path = None
    checkpoint_frequency = -1
    remove_tiles = False
    
    use_yolo_inference = False
    
    if not use_yolo_inference:
        
        yolo_inference_options = None
        
    else:
        
        yolo_inference_options = YoloInferenceOptions()
        yolo_inference_options.yolo_working_folder = os.path.expanduser('~/git/yolov5')
                    
    run_tiled_inference(model_file, image_folder, tiling_folder, output_file,
                            tile_size_x=tile_size_x, tile_size_y=tile_size_y, 
                            tile_overlap=tile_overlap,
                            checkpoint_path=checkpoint_path, 
                            checkpoint_frequency=checkpoint_frequency, 
                            remove_tiles=remove_tiles, 
                            yolo_inference_options=yolo_inference_options)
    
    
    #%% Run tiled inference (generate a command)
    
    import os
    
    model_file = os.path.expanduser('~/models/camera_traps/megadetector/md_v5.0.0/md_v5a.0.0.pt')
    image_folder = os.path.expanduser('~/data/KRU-test')
    tiling_folder = os.path.expanduser('~/tmp/tiling-test')
    output_file = os.path.expanduser('~/tmp/KRU-test-tiled.json')
    tile_size = [5152,3968]
    tile_overlap = 0.8
    
    cmd = f'python run_tiled_inference.py {model_file} {image_folder} {tiling_folder} {output_file} ' + \
          f'--tile_overlap {tile_overlap} --no_remove_tiles --tile_size_x {tile_size[0]} --tile_size_y {tile_size[1]}'
    
    print(cmd)
    import clipboard; clipboard.copy(cmd)
    
    
    #%% Preview tiled inference
    
    from api.batch_processing.postprocessing.postprocess_batch_results import (
        PostProcessingOptions, process_batch_results)

    options = PostProcessingOptions()
    options.image_base_dir = image_folder
    options.include_almost_detections = True
    options.num_images_to_sample = None
    options.confidence_threshold = 0.2
    options.almost_detection_confidence_threshold = options.confidence_threshold - 0.05
    options.ground_truth_json_file = None
    options.separate_detections_by_category = True
    # options.sample_seed = 0

    options.parallelize_rendering = True
    options.parallelize_rendering_n_cores = 10
    options.parallelize_rendering_with_threads = False

    preview_base = os.path.join(tiling_folder,'preview')
    os.makedirs(preview_base, exist_ok=True)

    print('Processing post-RDE to {}'.format(preview_base))

    options.api_output_file = output_file
    options.output_dir = preview_base
    ppresults = process_batch_results(options)
    html_output_file = ppresults.output_html_file

    path_utils.open_file(html_output_file)
    
    
#%% Command-line driver

import sys,argparse

def main():
            
    parser = argparse.ArgumentParser(
        description='Chop a folder of images up into tiles, run MD on the tiles, and stitch the results together')
    parser.add_argument(
        'model_file',
        help='Path to detector model file (.pb or .pt)')
    parser.add_argument(
        'image_folder',
        help='Folder containing images for inference (always recursive)')
    parser.add_argument(
        'tiling_folder',
        help='Temporary folder where tiles and intermediate results will be stored')
    parser.add_argument(
        'output_file',
        help='Path to output JSON results file, should end with a .json extension')
    parser.add_argument(
        '--no_remove_tiles',
        action='store_true',
        help='Tiles are removed by default; this option suppresses tile deletion')    
    parser.add_argument(
        '--tile_size_x',
        type=int,
        default=default_tile_size[0],
        help=('Tile width (defaults to {})'.format(default_tile_size[0])))
    parser.add_argument(
        '--tile_size_y',
        type=int,
        default=default_tile_size[0],
        help=('Tile height (defaults to {})'.format(default_tile_size[1])))
    parser.add_argument(
        '--tile_overlap',
        type=float,
        default=default_patch_overlap,
        help=('Overlap between tiles [0,1] (defaults to {})'.format(default_patch_overlap)))
        
    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    assert os.path.exists(args.model_file), \
        'detector file {} does not exist'.format(args.model_file)
    
    if os.path.exists(args.output_file):
        print('Warning: output_file {} already exists and will be overwritten'.format(
            args.output_file))

    remove_tiles = (not args.no_remove_tiles)

    run_tiled_inference(args.model_file, args.image_folder, args.tiling_folder, args.output_file,
                        tile_size_x=args.tile_size_x, tile_size_y=args.tile_size_y, 
                        tile_overlap=args.tile_overlap,
                        remove_tiles=remove_tiles)
        
if __name__ == '__main__':
    main()



########
# 
# visualization_utils.py
# 
# Core rendering functions shared across visualization scripts
#
########

#%% Constants and imports

from io import BytesIO
from typing import Union
import time

import matplotlib.pyplot as plt
import numpy as np
import requests
from PIL import Image, ImageFile, ImageFont, ImageDraw

from data_management.annotations import annotation_constants
from data_management.annotations.annotation_constants import (
    detector_bbox_category_id_to_name)  # here id is int

ImageFile.LOAD_TRUNCATED_IMAGES = True

IMAGE_ROTATIONS = {
    3: 180,
    6: 270,
    8: 90
}

TEXTALIGN_LEFT = 0
TEXTALIGN_RIGHT = 1

# convert category ID from int to str
DEFAULT_DETECTOR_LABEL_MAP = {
    str(k): v for k, v in detector_bbox_category_id_to_name.items()
}

# Retry on blob storage read failures
n_retries = 10
retry_sleep_time = 0.01
error_names_for_retry = ['ConnectionError']

DEFAULT_BOX_THICKNESS = 4
DEFAULT_LABEL_FONT_SIZE = 16


#%% Functions

def open_image(input_file: Union[str, BytesIO]) -> Image:
    """
    Opens an image in binary format using PIL.Image and converts to RGB mode.
    
    Supports local files or URLs.

    This operation is lazy; image will not be actually loaded until the first
    operation that needs to load it (for example, resizing), so file opening
    errors can show up later.

    Args:
        input_file: str or BytesIO, either a path to an image file (anything
            that PIL can open), or an image as a stream of bytes

    Returns:
        A PIL image object in RGB mode
    """
    
    if (isinstance(input_file, str)
            and input_file.startswith(('http://', 'https://'))):
        try:
            response = requests.get(input_file)
        except Exception as e:
            print(f'Error retrieving image {input_file}: {e}')
            success = False
            if e.__class__.__name__ in error_names_for_retry:
                for i_retry in range(0,n_retries):
                    try:
                        time.sleep(retry_sleep_time)
                        response = requests.get(input_file)        
                    except Exception as e:
                        print(f'Error retrieving image {input_file} on retry {i_retry}: {e}')
                        continue
                    print('Succeeded on retry {}'.format(i_retry))
                    success = True
                    break
            if not success:
                raise
        try:
            image = Image.open(BytesIO(response.content))
        except Exception as e:
            print(f'Error opening image {input_file}: {e}')
            raise

    else:
        image = Image.open(input_file)
    if image.mode not in ('RGBA', 'RGB', 'L', 'I;16'):
        raise AttributeError(
            f'Image {input_file} uses unsupported mode {image.mode}')
    if image.mode == 'RGBA' or image.mode == 'L':
        # PIL.Image.convert() returns a converted copy of this image
        image = image.convert(mode='RGB')

    # Alter orientation as needed according to EXIF tag 0x112 (274) for Orientation
    #
    # https://gist.github.com/dangtrinhnt/a577ece4cbe5364aad28
    # https://www.media.mit.edu/pia/Research/deepview/exif.html
    #
    try:
        exif = image._getexif()
        orientation: int = exif.get(274, None)  # 274 is the key for the Orientation field
        if orientation is not None and orientation in IMAGE_ROTATIONS:
            image = image.rotate(IMAGE_ROTATIONS[orientation], expand=True)  # returns a rotated copy
    except Exception:
        pass

    return image


def exif_preserving_save(pil_image,output_file):
    """
    Save [pil_image] to [output_file], making a moderate attempt to preserve EXIF
    data and JPEG quality.  Neither is guaranteed.
    
    Also see:
    
    https://discuss.dizzycoding.com/determining-jpg-quality-in-python-pil/
     
    ...for more ways to preserve jpeg quality if quality='keep' doesn't do the trick.
    """
    
    # Read EXIF metadata
    exif = pil_image.info['exif'] if ('exif' in pil_image.info) else None
    
    # Write output with EXIF metadata if available, and quality='keep' if this is a JPEG
    # image.  Unfortunately, neither parameter likes "None", so we get a slightly
    # icky cascade of if's here.
    if exif is not None:
        if pil_image.format == "JPEG":
            pil_image.save(output_file, exif=exif, quality='keep')
        else:
            pil_image.save(output_file, exif=exif)
    else:
        if pil_image.format == "JPEG":            
            pil_image.save(output_file, quality='keep')
        else:
            pil_image.save(output_file)
            
            
def load_image(input_file: Union[str, BytesIO]) -> Image:
    """
    Loads the image at input_file as a PIL Image into memory.

    Image.open() used in open_image() is lazy and errors will occur downstream
    if not explicitly loaded.

    Args:
        input_file: str or BytesIO, either a path to an image file (anything
            that PIL can open), or an image as a stream of bytes

    Returns: PIL.Image.Image, in RGB mode
    """
    
    image = open_image(input_file)
    image.load()
    return image


def resize_image(image, target_width, target_height=-1):
    """
    Resizes a PIL image object to the specified width and height; does not resize
    in place. If either width or height are -1, resizes with aspect ratio preservation.
    If both are -1, returns the original image (does not copy in this case).
    """

    # Null operation
    if target_width == -1 and target_height == -1:
        return image

    elif target_width == -1 or target_height == -1:

        # Aspect ratio as width over height
        # ar = w / h
        aspect_ratio = image.size[0] / image.size[1]

        if target_width != -1:
            # h = w / ar
            target_height = int(target_width / aspect_ratio)
        else:
            # w = ar * h
            target_width = int(aspect_ratio * target_height)

    # This parameter changed between Pillow versions 9 and 10, and for a bit, I'd like to
    # support both.
    try:
        resized_image = image.resize((target_width, target_height), Image.ANTIALIAS)
    except:
        resized_image = image.resize((target_width, target_height), Image.Resampling.LANCZOS)
        
    return resized_image


def show_images_in_a_row(images):

    num = len(images)
    assert num > 0

    if isinstance(images[0], str):
        images = [Image.open(img) for img in images]

    fig, axarr = plt.subplots(1, num, squeeze=False)  # number of rows, number of columns
    fig.set_size_inches((num * 5, 25))  # each image is 2 inches wide
    for i, img in enumerate(images):
        axarr[0, i].set_axis_off()
        axarr[0, i].imshow(img)
    return fig


# The following three functions are modified versions of those at:
#
# https://github.com/tensorflow/models/blob/master/research/object_detection/utils/visualization_utils.py

DEFAULT_COLORS = [
    'AliceBlue', 'Red', 'RoyalBlue', 'Gold', 'Chartreuse', 'Aqua', 'Azure',
    'Beige', 'Bisque', 'BlanchedAlmond', 'BlueViolet', 'BurlyWood', 'CadetBlue',
    'AntiqueWhite', 'Chocolate', 'Coral', 'CornflowerBlue', 'Cornsilk', 'Crimson',
    'Cyan', 'DarkCyan', 'DarkGoldenRod', 'DarkGrey', 'DarkKhaki', 'DarkOrange',
    'DarkOrchid', 'DarkSalmon', 'DarkSeaGreen', 'DarkTurquoise', 'DarkViolet',
    'DeepPink', 'DeepSkyBlue', 'DodgerBlue', 'FireBrick', 'FloralWhite',
    'ForestGreen', 'Fuchsia', 'Gainsboro', 'GhostWhite', 'GoldenRod',
    'Salmon', 'Tan', 'HoneyDew', 'HotPink', 'IndianRed', 'Ivory', 'Khaki',
    'Lavender', 'LavenderBlush', 'LawnGreen', 'LemonChiffon', 'LightBlue',
    'LightCoral', 'LightCyan', 'LightGoldenRodYellow', 'LightGray', 'LightGrey',
    'LightGreen', 'LightPink', 'LightSalmon', 'LightSeaGreen', 'LightSkyBlue',
    'LightSlateGray', 'LightSlateGrey', 'LightSteelBlue', 'LightYellow', 'Lime',
    'LimeGreen', 'Linen', 'Magenta', 'MediumAquaMarine', 'MediumOrchid',
    'MediumPurple', 'MediumSeaGreen', 'MediumSlateBlue', 'MediumSpringGreen',
    'MediumTurquoise', 'MediumVioletRed', 'MintCream', 'MistyRose', 'Moccasin',
    'NavajoWhite', 'OldLace', 'Olive', 'OliveDrab', 'Orange', 'OrangeRed',
    'Orchid', 'PaleGoldenRod', 'PaleGreen', 'PaleTurquoise', 'PaleVioletRed',
    'PapayaWhip', 'PeachPuff', 'Peru', 'Pink', 'Plum', 'PowderBlue', 'Purple',
    'RosyBrown', 'Aquamarine', 'SaddleBrown', 'Green', 'SandyBrown',
    'SeaGreen', 'SeaShell', 'Sienna', 'Silver', 'SkyBlue', 'SlateBlue',
    'SlateGray', 'SlateGrey', 'Snow', 'SpringGreen', 'SteelBlue', 'GreenYellow',
    'Teal', 'Thistle', 'Tomato', 'Turquoise', 'Violet', 'Wheat', 'White',
    'WhiteSmoke', 'Yellow', 'YellowGreen'
]


def crop_image(detections, image, confidence_threshold=0.15, expansion=0):
    """
    Crops detections above *confidence_threshold* from the PIL image *image*,
    returning a list of PIL images.

    *detections* should be a list of dictionaries with keys 'conf' and 'bbox';
    see bbox format description below.  Normalized, [x,y,w,h], upper-left-origin.

    *expansion* specifies a number of pixels to include on each side of the box.
    """

    ret_images = []

    for detection in detections:

        score = float(detection['conf'])

        if score >= confidence_threshold:

            x1, y1, w_box, h_box = detection['bbox']
            ymin,xmin,ymax,xmax = y1, x1, y1 + h_box, x1 + w_box

            # Convert to pixels so we can use the PIL crop() function
            im_width, im_height = image.size
            (left, right, top, bottom) = (xmin * im_width, xmax * im_width,
                                          ymin * im_height, ymax * im_height)

            if expansion > 0:
                left -= expansion
                right += expansion
                top -= expansion
                bottom += expansion

            # PIL's crop() does surprising things if you provide values outside of
            # the image, clip inputs
            left = max(left,0); right = max(right,0)
            top = max(top,0); bottom = max(bottom,0)

            left = min(left,im_width-1); right = min(right,im_width-1)
            top = min(top,im_height-1); bottom = min(bottom,im_height-1)

            ret_images.append(image.crop((left, top, right, bottom)))

        # ...if this detection is above threshold

    # ...for each detection

    return ret_images


def render_detection_bounding_boxes(detections, image,
                                    label_map={}, 
                                    classification_label_map=None, 
                                    confidence_threshold=0.15, thickness=DEFAULT_BOX_THICKNESS, expansion=0,
                                    classification_confidence_threshold=0.3,
                                    max_classifications=3,
                                    colormap=DEFAULT_COLORS,
                                    textalign=TEXTALIGN_LEFT,
                                    label_font_size=DEFAULT_LABEL_FONT_SIZE,
                                    custom_strings=None):
    """
    Renders bounding boxes, label, and confidence on an image if confidence is above the threshold.

    Boxes are in the format that's output from the batch processing API.

    Renders classification labels if present.

    Args:

        detections: detections on the image, example content:
            [
                {
                    "category": "2",
                    "conf": 0.996,
                    "bbox": [
                        0.0,
                        0.2762,
                        0.1234,
                        0.2458
                    ]
                }
            ]

            ...where the bbox coordinates are [x, y, box_width, box_height].

            (0, 0) is the upper-left.  Coordinates are normalized.

            Supports classification results, if *detections* has the format
            [
                {
                    "category": "2",
                    "conf": 0.996,
                    "bbox": [
                        0.0,
                        0.2762,
                        0.1234,
                        0.2458
                    ]
                    "classifications": [
                        ["3", 0.901],
                        ["1", 0.071],
                        ["4", 0.025]
                    ]
                }
            ]

        image: PIL.Image object

        label_map: optional, mapping the numerical label to a string name. The type of the numerical label
            (default string) needs to be consistent with the keys in label_map; no casting is carried out.
            If this is None, no labels are shown.

        classification_label_map: optional, mapping of the string class labels to the actual class names.
            The type of the numerical label (default string) needs to be consistent with the keys in
            label_map; no casting is carried out.  If this is None, no classification labels are shown.

        confidence_threshold: optional, threshold above which the bounding box is rendered.
        
        thickness: line thickness in pixels. Default value is 4.
        
        expansion: number of pixels to expand bounding boxes on each side.  Default is 0.
        
        classification_confidence_threshold: confidence above which classification result is retained.
        
        max_classifications: maximum number of classification results retained for one image.
        
        custom_strings: optional set of strings to append to detection labels, should have the
        same length as [detections].  Appended before classification labels, if classification
        data is provided.

    image is modified in place.
    """

    if custom_strings is not None:
        assert len(custom_strings) == len(detections), \
            '{} custom strings provided for {} detections'.format(
                len(custom_strings),len(detections))
            
    display_boxes = []
    
    # list of lists, one list of strings for each bounding box (to accommodate multiple labels)
    display_strs = []  
    
    # for color selection
    classes = []  

    for i_detection,detection in enumerate(detections):

        score = detection['conf']
        
        # Always render objects with a confidence of "None", this is typically used
        # for ground truth data.        
        if score is None or score >= confidence_threshold:
            
            x1, y1, w_box, h_box = detection['bbox']
            display_boxes.append([y1, x1, y1 + h_box, x1 + w_box])
            clss = detection['category']
            
            # {} is the default, which means "show labels with no mapping", so don't use "if label_map" here
            # if label_map:
            if label_map is not None:
                label = label_map[clss] if clss in label_map else clss
                if score is not None:
                    displayed_label = ['{}: {}%'.format(label, round(100 * score))]
                else:
                    displayed_label = ['{}'.format(label)]
            else:
                displayed_label = ''

            if custom_strings is not None:
                custom_string = custom_strings[i_detection]
                if custom_string is not None and len(custom_string) > 0:
                    if isinstance(displayed_label,str):
                        displayed_label += ' ' + custom_string
                    else:
                        assert len(displayed_label) == 1
                        displayed_label[0] += ' ' + custom_string
                                    
            if 'classifications' in detection:

                # To avoid duplicate colors with detection-only visualization, offset
                # the classification class index by the number of detection classes
                clss = annotation_constants.NUM_DETECTOR_CATEGORIES + int(detection['classifications'][0][0])
                classifications = detection['classifications']
                if len(classifications) > max_classifications:
                    classifications = classifications[0:max_classifications]
                    
                for classification in classifications:
                    
                    classification_conf = classification[1]
                    if classification_conf is not None and \
                        classification_conf < classification_confidence_threshold:
                        continue
                    class_key = classification[0]
                    if (classification_label_map is not None) and (class_key in classification_label_map):
                        class_name = classification_label_map[class_key]
                    else:
                        class_name = class_key
                    if classification_conf is not None:
                        displayed_label += ['{}: {:5.1%}'.format(class_name.lower(), classification_conf)]
                    else:
                        displayed_label += ['{}'.format(class_name.lower())]
                    
                # ...for each classification

            # ...if we have classification results
                        
            display_strs.append(displayed_label)
            classes.append(clss)

        # ...if the confidence of this detection is above threshold

    # ...for each detection
    
    display_boxes = np.array(display_boxes)

    draw_bounding_boxes_on_image(image, display_boxes, classes,
                                 display_strs=display_strs, thickness=thickness, 
                                 expansion=expansion, colormap=colormap, textalign=textalign,
                                 label_font_size=label_font_size)


def draw_bounding_boxes_on_image(image,
                                 boxes,
                                 classes,
                                 thickness=DEFAULT_BOX_THICKNESS,
                                 expansion=0,
                                 display_strs=None,
                                 colormap=DEFAULT_COLORS,
                                 textalign=TEXTALIGN_LEFT,
                                 label_font_size=DEFAULT_LABEL_FONT_SIZE):
    """
    Draws bounding boxes on an image.

    Args:
      image: a PIL.Image object.
      boxes: a 2 dimensional numpy array of [N, 4]: (ymin, xmin, ymax, xmax).
             The coordinates are in normalized format between [0, 1].
      classes: a list of ints or strings (that can be cast to ints) corresponding to the
               class labels of the boxes. This is only used for color selection.
      thickness: line thickness in pixels. Default value is 4.
      expansion: number of pixels to expand bounding boxes on each side.  Default is 0.
      display_strs: list of list of strings.
                             a list of strings for each bounding box.
                             The reason to pass a list of strings for a
                             bounding box is that it might contain
                             multiple labels.
    """

    boxes_shape = boxes.shape
    if not boxes_shape:
        return
    if len(boxes_shape) != 2 or boxes_shape[1] != 4:
        # print('Input must be of size [N, 4], but is ' + str(boxes_shape))
        return  # no object detection on this image, return
    for i in range(boxes_shape[0]):
        if display_strs:
            display_str_list = display_strs[i]
            draw_bounding_box_on_image(image,
                                       boxes[i, 0], boxes[i, 1], boxes[i, 2], boxes[i, 3],
                                       classes[i],
                                       thickness=thickness, expansion=expansion,
                                       display_str_list=display_str_list,
                                       colormap=colormap,
                                       textalign=textalign,
                                       label_font_size=label_font_size)


def draw_bounding_box_on_image(image,
                               ymin,
                               xmin,
                               ymax,
                               xmax,
                               clss=None,
                               thickness=DEFAULT_BOX_THICKNESS,
                               expansion=0,
                               display_str_list=(),
                               use_normalized_coordinates=True,
                               label_font_size=DEFAULT_LABEL_FONT_SIZE,
                               colormap=DEFAULT_COLORS,
                               textalign=TEXTALIGN_LEFT):
    """
    Adds a bounding box to an image.

    Bounding box coordinates can be specified in either absolute (pixel) or
    normalized coordinates by setting the use_normalized_coordinates argument.

    Each string in display_str_list is displayed on a separate line above the
    bounding box in black text on a rectangle filled with the input 'color'.
    If the top of the bounding box extends to the edge of the image, the strings
    are displayed below the bounding box.

    Args:
    image: a PIL.Image object.
    ymin: ymin of bounding box - upper left.
    xmin: xmin of bounding box.
    ymax: ymax of bounding box.
    xmax: xmax of bounding box.
    clss: str, the class of the object in this bounding box - will be cast to an int.
    thickness: line thickness. Default value is 4.
    expansion: number of pixels to expand bounding boxes on each side.  Default is 0.
    display_str_list: list of strings to display in box
        (each to be shown on its own line).
        use_normalized_coordinates: If True (default), treat coordinates
        ymin, xmin, ymax, xmax as relative to the image.  Otherwise treat
        coordinates as absolute.
    label_font_size: font size to attempt to load arial.ttf with
    """
    
    if clss is None:
        color = colormap[1]
    else:
        color = colormap[int(clss) % len(colormap)]

    draw = ImageDraw.Draw(image)
    im_width, im_height = image.size
    if use_normalized_coordinates:
        (left, right, top, bottom) = (xmin * im_width, xmax * im_width,
                                      ymin * im_height, ymax * im_height)
    else:
        (left, right, top, bottom) = (xmin, xmax, ymin, ymax)

    if expansion > 0:
        
        left -= expansion
        right += expansion
        top -= expansion
        bottom += expansion
        
        # Deliberately trimming to the width of the image only in the case where
        # box expansion is turned on.  There's not an obvious correct behavior here,
        # but the thinking is that if the caller provided an out-of-range bounding
        # box, they meant to do that, but at least in the eyes of the person writing
        # this comment, if you expand a box for visualization reasons, you don't want
        # to end up with part of a box.
        #
        # A slightly more sophisticated might check whether it was in fact the expansion
        # that made this box larger than the image, but this is the case 99.999% of the time
        # here, so that doesn't seem necessary.
        left = max(left,0); right = max(right,0)
        top = max(top,0); bottom = max(bottom,0)

        left = min(left,im_width-1); right = min(right,im_width-1)
        top = min(top,im_height-1); bottom = min(bottom,im_height-1)
        
    # ...if we need to expand boxes
    
    draw.line([(left, top), (left, bottom), (right, bottom),
               (right, top), (left, top)], width=thickness, fill=color)

    try:
        font = ImageFont.truetype('arial.ttf', label_font_size)
    except IOError:
        font = ImageFont.load_default()

    def get_text_size(font,s):

        # This is what we did w/Pillow 9
        # w,h = font.getsize(s)
        
        # I would *think* this would be the equivalent for Pillow 10
        # l,t,r,b = font.getbbox(s); w = r-l; h=b-t
        
        # ...but this actually produces the most similar results to Pillow 9
        # l,t,r,b = font.getbbox(s); w = r; h=b
        
        try:
            l,t,r,b = font.getbbox(s); w = r; h=b  
        except Exception:
            w,h = font.getsize(s)
        
        return w,h
    
    # If the total height of the display strings added to the top of the bounding
    # box exceeds the top of the image, stack the strings below the bounding box
    # instead of above.
    display_str_heights = [get_text_size(font,ds)[1] for ds in display_str_list]

    # Each display_str has a top and bottom margin of 0.05x.
    total_display_str_height = (1 + 2 * 0.05) * sum(display_str_heights)

    if top > total_display_str_height:
        text_bottom = top
    else:
        text_bottom = bottom + total_display_str_height

    # Reverse list and print from bottom to top.
    for display_str in display_str_list[::-1]:

        # Skip empty strings
        if len(display_str) == 0:
            continue
        
        text_width, text_height = get_text_size(font,display_str)
        
        text_left = left
        
        if textalign == TEXTALIGN_RIGHT:
            text_left = right - text_width
            
        margin = np.ceil(0.05 * text_height)

        draw.rectangle(
            [(text_left, text_bottom - text_height - 2 * margin), (text_left + text_width,
                                                              text_bottom)],
            fill=color)

        draw.text(
            (text_left + margin, text_bottom - text_height - margin),
            display_str,
            fill='black',
            font=font)

        text_bottom -= (text_height + 2 * margin)


def render_iMerit_boxes(boxes, classes, image,
                        label_map=annotation_constants.annotation_bbox_category_id_to_name):
    """
    Renders bounding boxes and their category labels on a PIL image.

    Args:
        boxes: bounding box annotations from iMerit, format is:
            [x_rel, y_rel, w_rel, h_rel] (rel = relative coords)
        classes: the class IDs of the predicted class of each box/object
        image: PIL.Image object to annotate on
        label_map: optional dict mapping classes to a string for display

    Returns:
        image will be altered in place
    """

    display_boxes = []
    
    # list of lists, one list of strings for each bounding box (to accommodate multiple labels)
    display_strs = []  
    
    for box, clss in zip(boxes, classes):
        if len(box) == 0:
            assert clss == 5
            continue
        x_rel, y_rel, w_rel, h_rel = box
        ymin, xmin = y_rel, x_rel
        ymax = ymin + h_rel
        xmax = xmin + w_rel

        display_boxes.append([ymin, xmin, ymax, xmax])

        if label_map:
            clss = label_map[int(clss)]
        display_strs.append([clss])

    display_boxes = np.array(display_boxes)
    draw_bounding_boxes_on_image(image, display_boxes, classes, display_strs=display_strs)


def render_megadb_bounding_boxes(boxes_info, image):
    """
    Args:
        boxes_info: list of dict, each dict represents a single detection
            {
                "category": "animal",
                "bbox": [
                    0.739,
                    0.448,
                    0.187,
                    0.198
                ]
            }
            where bbox coordinates are normalized [x_min, y_min, width, height]
        image: PIL.Image.Image, opened image
    """
    
    display_boxes = []
    display_strs = []
    classes = []  # ints, for selecting colors

    for b in boxes_info:
        x_min, y_min, w_rel, h_rel = b['bbox']
        y_max = y_min + h_rel
        x_max = x_min + w_rel
        display_boxes.append([y_min, x_min, y_max, x_max])
        display_strs.append([b['category']])
        classes.append(annotation_constants.detector_bbox_category_name_to_id[b['category']])

    display_boxes = np.array(display_boxes)
    draw_bounding_boxes_on_image(image, display_boxes, classes, display_strs=display_strs)


def render_db_bounding_boxes(boxes, classes, image, original_size=None,
                             label_map=None, thickness=DEFAULT_BOX_THICKNESS, expansion=0):
    """
    Render bounding boxes (with class labels) on [image].  This is a wrapper for
    draw_bounding_boxes_on_image, allowing the caller to operate on a resized image
    by providing the original size of the image; bboxes will be scaled accordingly.
    
    This function assumes that bounding boxes are in the COCO camera traps format,
    with absolute coordinates.
    """

    display_boxes = []
    display_strs = []

    if original_size is not None:
        image_size = original_size
    else:
        image_size = image.size

    img_width, img_height = image_size

    for box, clss in zip(boxes, classes):

        x_min_abs, y_min_abs, width_abs, height_abs = box[0:4]

        ymin = y_min_abs / img_height
        ymax = ymin + height_abs / img_height

        xmin = x_min_abs / img_width
        xmax = xmin + width_abs / img_width

        display_boxes.append([ymin, xmin, ymax, xmax])

        if label_map:
            clss = label_map[int(clss)]
            
        # need to be a string here because PIL needs to iterate through chars
        display_strs.append([str(clss)])  

    display_boxes = np.array(display_boxes)
    draw_bounding_boxes_on_image(image, display_boxes, classes, display_strs=display_strs,
                                 thickness=thickness, expansion=expansion)


def draw_bounding_boxes_on_file(input_file, output_file, detections, confidence_threshold=0.0,
                                detector_label_map=DEFAULT_DETECTOR_LABEL_MAP,
                                thickness=DEFAULT_BOX_THICKNESS, expansion=0,
                                colormap=DEFAULT_COLORS,
                                custom_strings=None):
    """
    Render detection bounding boxes on an image loaded from file, writing the results to a
    new image file.
    
    "detections" is in the API results format:
        
    [{"category": "2","conf": 0.996,"bbox": [0.0,0.2762,0.1234,0.2458]}]
    
    ...where the bbox is:
        
    [x_min, y_min, width_of_box, height_of_box]
    
    Normalized, with the origin at the upper-left.
    
    detector_label_map is a dict mapping category IDs to strings.
    
    custom_strings: optional set of strings to append to detection labels, should have the
    same length as [detections].  Appended before classification labels, if classification
    data is provided.
    """
    
    image = open_image(input_file)

    render_detection_bounding_boxes(
            detections, image, label_map=detector_label_map,
            confidence_threshold=confidence_threshold,
            thickness=thickness,expansion=expansion,colormap=colormap,
            custom_strings=custom_strings)

    image.save(output_file)


def draw_db_boxes_on_file(input_file, output_file, boxes, classes=None, 
                          label_map=None, thickness=DEFAULT_BOX_THICKNESS, expansion=0):
    """
    Render COCO bounding boxes (in absolute coordinates) on an image loaded from file, writing the
    results to a new image file.

    classes is a list of integer category IDs.
    
    detector_label_map is a dict mapping category IDs to strings.
    """
    
    image = open_image(input_file)

    if classes is None:
        classes = [0] * len(boxes)
        
    render_db_bounding_boxes(boxes, classes, image, original_size=None,
                                 label_map=label_map, thickness=thickness, expansion=expansion)

    image.save(output_file)
    
    


########
#
# write_html_image_list.py
#
# Given a list of image file names, writes an HTML file that
# shows all those images, with optional one-line headers above each.
#
# Each "filename" can also be a dict with elements 'filename','title',
# 'imageStyle','textStyle', 'linkTarget'
#
# Strips directory information away if options.makeRelative == 1.
#
# Tries to convert absolute to relative paths if options.makeRelative == 2.
#
########

#%% Constants and imports

import os
import math

from md_utils import path_utils


#%% write_html_image_list

def write_html_image_list(filename=None,images=None,options=None):
    """
    filename: the output file
    
    image: a list of image filenames or dictionaries with one or more of the following fields:
        
        filename
        imageStyle
        textStyle
        title
        linkTarget
        
    options: a dict with one or more of the following fields:
        
        fHtml
        makeRelative
        headerHtml
        trailerHtml
        defaultTextStyle
        defaultImageStyle
        maxFiguresPerHtmlFile
        
    """
    
    # returns an options struct
    if options is None:
        options = {}
        
    if 'fHtml' not in options:
        options['fHtml'] = -1
    
    if 'makeRelative' not in options:
        options['makeRelative'] = 0
    
    if 'headerHtml' not in options or options['headerHtml'] is None:
        options['headerHtml'] = ''        
    
    if 'trailerHtml' not in options or options['trailerHtml'] is None:
        options['trailerHtml'] = ''    
    
    if 'defaultTextStyle' not in options or options['defaultTextStyle'] is None:
        options['defaultTextStyle'] = \
        "font-family:calibri,verdana,arial;font-weight:bold;font-size:150%;text-align:left;margin:0px;"

    if 'defaultImageStyle' not in options or options['defaultImageStyle'] is None:
        options['defaultImageStyle'] = \
        "margin:0px;margin-top:5px;margin-bottom:5px;"
        
    # Possibly split the html output for figures into multiple files; Chrome gets sad with
    # thousands of images in a single tab.        
    if 'maxFiguresPerHtmlFile' not in options or options['maxFiguresPerHtmlFile'] is None:
        options['maxFiguresPerHtmlFile'] = math.inf    
    
    if filename is None:
        return options
    
    # images may be a list of images or a list of image/style/title dictionaries, 
    # enforce that it's the latter to simplify downstream code
    for iImage,imageInfo in enumerate(images):
        if isinstance(imageInfo,str):
            imageInfo = {'filename':imageInfo}
        if 'filename' not in imageInfo:
            imageInfo['filename'] = ''
        if 'imageStyle' not in imageInfo:
            imageInfo['imageStyle'] = options['defaultImageStyle']
        if 'title' not in imageInfo:
            imageInfo['title'] = ''
        if 'linkTarget' not in imageInfo:
            imageInfo['linkTarget'] = ''
        if 'textStyle' not in imageInfo:
            textStyle = options['defaultTextStyle']
            imageInfo['textStyle'] = options['defaultTextStyle']
        images[iImage] = imageInfo            
    
    # Remove leading directory information from filenames if requested
    if options['makeRelative'] == 1:
        
        for iImage in range(0,len(images)):
            _,n,e = path_utils.fileparts(images[iImage]['filename'])
            images[iImage]['filename'] = n + e
        
    elif options['makeRelative'] == 2:
        
        baseDir,_,_ = path_utils.fileparts(filename)
        if len(baseDir) > 1 and baseDir[-1] != '\\':
            baseDir = baseDir + '\\'
        
        for iImage in range(0,len(images)):
            fn = images[iImage]['filename']
            fn = fn.replace(baseDir,'')
            images[iImage]['filename'] = fn        
    
    nImages = len(images)
    
    # If we need to break this up into multiple files...
    if nImages > options['maxFiguresPerHtmlFile']:
    
        # You can't supply your own file handle in this case
        if options['fHtml'] != -1:
            raise ValueError(
                    'You can''t supply your own file handle if we have to page the image set')
        
        figureFileStartingIndices = list(range(0,nImages,options['maxFiguresPerHtmlFile']))

        assert len(figureFileStartingIndices) > 1
        
        # Open the meta-output file
        fMeta = open(filename,'w')
        
        # Write header stuff
        fMeta.write('<html><body>\n')    
        fMeta.write(options['headerHtml'])        
        fMeta.write('<table border = 0 cellpadding = 2>\n')
        
        for startingIndex in figureFileStartingIndices:
            
            iStart = startingIndex
            iEnd = startingIndex+options['maxFiguresPerHtmlFile']-1;
            if iEnd >= nImages:
                iEnd = nImages-1
            
            trailer = 'image_{:05d}_{:05d}'.format(iStart,iEnd)
            localFiguresHtmlFilename = path_utils.insert_before_extension(filename,trailer)
            fMeta.write('<tr><td>\n')
            fMeta.write('<p style="padding-bottom:0px;margin-bottom:0px;text-align:left;font-family:''segoe ui'',calibri,arial;font-size:100%;text-decoration:none;font-weight:bold;">')
            fMeta.write('<a href="{}">Figures for images {} through {}</a></p></td></tr>\n'.format(
                os.path.basename(localFiguresHtmlFilename),iStart,iEnd))
            
            localImages = images[iStart:iEnd+1]
            
            localOptions = options.copy();
            localOptions['headerHtml'] = '';
            localOptions['trailerHtml'] = '';
            
            # Make a recursive call for this image set
            write_html_image_list(localFiguresHtmlFilename,localImages,localOptions)
            
        # ...for each page of images
        
        fMeta.write('</table></body>\n')
        fMeta.write(options['trailerHtml'])
        fMeta.write('</html>\n')
        fMeta.close()
        
        return options
        
    # ...if we have to make multiple sub-pages
        
    bCleanupFile = False
    
    if options['fHtml'] == -1:
        bCleanupFile = True;
        fHtml = open(filename,'w')
    else:
        fHtml = options['fHtml']
        
    fHtml.write('<html><body>\n')
    
    fHtml.write(options['headerHtml'])
    
    # Write out images
    for iImage,image in enumerate(images):
        
        title = image['title']
        imageStyle = image['imageStyle']
        textStyle = image['textStyle']
        filename = image['filename']
        linkTarget = image['linkTarget']
        
        # Remove unicode characters
        title = title.encode('ascii','ignore').decode('ascii')
        filename = filename.encode('ascii','ignore').decode('ascii')
        
        if len(title) > 0:       
            fHtml.write(
                    '<p style="{}">{}</p>\n'\
                    .format(textStyle,title))            

        if len(linkTarget) > 0:
            fHtml.write('<a href="{}">'.format(linkTarget))
            # imageStyle.append(';border:0px;')
        
        fHtml.write('<img src="{}" style="{}">\n'.format(filename,imageStyle))
        
        if len(linkTarget) > 0:
            fHtml.write('</a>')
            
        if iImage != len(images)-1:
            fHtml.write('<br/>')             
            
    # ...for each image we need to write
    
    fHtml.write(options['trailerHtml'])
    
    fHtml.write('</body></html>\n')
    
    if bCleanupFile:
        fHtml.close()    

# ...function



########
# 
# path_utils.py
#
# Miscellaneous useful utils for path manipulation, things that could *almost*
# be in os.path, but aren't.
#
########

#%% Imports and constants

import glob
import ntpath
import os
import posixpath
import string
import json
import unicodedata
import zipfile

from zipfile import ZipFile
from datetime import datetime
from typing import Container, Iterable, List, Optional, Tuple, Sequence
from multiprocessing.pool import ThreadPool
from tqdm import tqdm

IMG_EXTENSIONS = ('.jpg', '.jpeg', '.gif', '.png', '.tif', '.tiff', '.bmp')

VALID_FILENAME_CHARS = f"~-_.() {string.ascii_letters}{string.digits}"
SEPARATOR_CHARS = r":\/"
VALID_PATH_CHARS = VALID_FILENAME_CHARS + SEPARATOR_CHARS
CHAR_LIMIT = 255


#%% General path functions

def recursive_file_list(base_dir, convert_slashes=True, return_relative_paths=False):
    """
    Enumerate files (not directories) in [base_dir], optionally converting
    \ to /
    """
    
    all_files = []

    for root, _, filenames in os.walk(base_dir):
        for filename in filenames:
            full_path = os.path.join(root, filename)
            if convert_slashes:
                full_path = full_path.replace('\\', '/')
            all_files.append(full_path)

    if return_relative_paths:
        all_files = [os.path.relpath(fn,base_dir) for fn in all_files]
        
    all_files = sorted(all_files)
    return all_files


def split_path(path: str) -> List[str]:
    """
    Splits [path] into all its constituent tokens.

    Non-recursive version of:
    http://nicks-liquid-soapbox.blogspot.com/2011/03/splitting-path-to-list-in-python.html

    Examples
    >>> split_path(r'c:\dir\subdir\file.txt')
    ['c:\\', 'dir', 'subdir', 'file.txt']
    >>> split_path('/dir/subdir/file.jpg')
    ['/', 'dir', 'subdir', 'file.jpg']
    >>> split_path('c:\\')
    ['c:\\']
    >>> split_path('/')
    ['/']
    """
    
    parts = []
    while True:
        # ntpath seems to do the right thing for both Windows and Unix paths
        head, tail = ntpath.split(path)
        if head == '' or head == path:
            break
        parts.append(tail)
        path = head
    parts.append(head or tail)
    return parts[::-1]  # reverse


def fileparts(path: str) -> Tuple[str, str, str]:
    """
    Breaks down a path into the directory path, filename, and extension.

    Note that the '.' lives with the extension, and separators are removed.

    Examples
    >>> fileparts('file')
    ('', 'file', '')
    >>> fileparts(r'c:\dir\file.jpg')
    ('c:\\dir', 'file', '.jpg')
    >>> fileparts('/dir/subdir/file.jpg')
    ('/dir/subdir', 'file', '.jpg')

    Returns:
        p: str, directory path
        n: str, filename without extension
        e: str, extension including the '.'
    """
    
    # ntpath seems to do the right thing for both Windows and Unix paths
    p = ntpath.dirname(path)
    basename = ntpath.basename(path)
    n, e = ntpath.splitext(basename)
    return p, n, e


def insert_before_extension(filename: str, s: str = '', separator='.') -> str:
    """
    Insert string [s] before the extension in [filename], separated with [separator].

    If [s] is empty, generates a date/timestamp. If [filename] has no extension,
    appends [s].

    Examples
    >>> insert_before_extension('/dir/subdir/file.ext', 'insert')
    '/dir/subdir/file.insert.ext'
    >>> insert_before_extension('/dir/subdir/file', 'insert')
    '/dir/subdir/file.insert'
    >>> insert_before_extension('/dir/subdir/file')
    '/dir/subdir/file.2020.07.20.10.54.38'
    """
    
    assert len(filename) > 0
    if len(s) == 0:
        s = datetime.now().strftime('%Y.%m.%d.%H.%M.%S')
    name, ext = os.path.splitext(filename)
    return f'{name}{separator}{s}{ext}'


def top_level_folder(p: str, windows: Optional[bool] = None) -> str:
    """
    Gets the top-level folder from path [p].

    This function behaves differently for Windows vs. Unix paths. Set
    windows=True if [p] is a Windows path. Set windows=None (default) to treat
    [p] as a native system path.

    On Windows, will use the top-level folder that isn't the drive.
    >>> top_level_folder(r'c:\blah\foo')
    'c:\blah'

    On Unix, does not include the leaf node.
    >>> top_level_folder('/blah/foo')
    '/blah'
    """
    
    if p == '':
        return ''

    default_lib = os.path  # save default os.path
    if windows is not None:
        os.path = ntpath if windows else posixpath

    # Path('/blah').parts is ('/', 'blah')
    parts = split_path(p)

    drive = os.path.splitdrive(p)[0]
    if len(parts) > 1 and (
            parts[0] == drive
            or parts[0] == drive + '/'
            or parts[0] == drive + '\\'
            or parts[0] in ['\\', '/']):
        result = os.path.join(parts[0], parts[1])
    else:
        result = parts[0]

    os.path = default_lib  # restore default os.path
    return result


def safe_create_link(link_exists,link_new):
    """
    Create a symlink at link_new pointing to link_exists.
    
    If link_new already exists, make sure it's a link (not a file),
    and if it has a different target than link_exists, remove and re-create
    it.
    
    Errors of link_new already exists but it's not a link.
    """    
    if os.path.exists(link_new) or os.path.islink(link_new):
        assert os.path.islink(link_new)
        if not os.readlink(link_new) == link_exists:
            os.remove(link_new)
            os.symlink(link_exists,link_new)
    else:
        os.symlink(link_exists,link_new)
        

#%% Image-related path functions

def is_image_file(s: str, img_extensions: Container[str] = IMG_EXTENSIONS
                  ) -> bool:
    """
    Checks a file's extension against a hard-coded set of image file
    extensions.
    
    Does not check whether the file exists, only determines whether the filename
    implies it's an image file.
    """
    
    ext = os.path.splitext(s)[1]
    return ext.lower() in img_extensions


def find_image_strings(strings: Iterable[str]) -> List[str]:
    """
    Given a list of strings that are potentially image file names, looks for
    strings that actually look like image file names (based on extension).
    """
    
    return [s for s in strings if is_image_file(s)]


def find_images(dirname: str, recursive: bool = False, return_relative_paths: bool = False) -> List[str]:
    """
    Finds all files in a directory that look like image file names. Returns
    absolute paths unless return_relative_paths is set.    
    """
    
    if recursive:
        strings = glob.glob(os.path.join(dirname, '**', '*.*'), recursive=True)
    else:
        strings = glob.glob(os.path.join(dirname, '*.*'))
    
    image_files = find_image_strings(strings)
    
    if return_relative_paths:
        image_files = [os.path.relpath(fn,dirname) for fn in image_files]
    
    image_files = sorted(image_files)
    return image_files


#%% Filename cleaning functions

def clean_filename(filename: str, allow_list: str = VALID_FILENAME_CHARS,
                   char_limit: int = CHAR_LIMIT, force_lower: bool = False) -> str:
    """
    Removes non-ASCII and other invalid filename characters (on any
    reasonable OS) from a filename, then trims to a maximum length.

    Does not allow :\/, use clean_path if you want to preserve those.

    Adapted from
    https://gist.github.com/wassname/1393c4a57cfcbf03641dbc31886123b8
    """
    
    # keep only valid ascii chars
    cleaned_filename = (unicodedata.normalize('NFKD', filename)
                        .encode('ASCII', 'ignore').decode())

    # keep only allow-listed chars
    cleaned_filename = ''.join([c for c in cleaned_filename if c in allow_list])
    if char_limit is not None:
        cleaned_filename = cleaned_filename[:char_limit]
    if force_lower:
        cleaned_filename = cleaned_filename.lower()
    return cleaned_filename


def clean_path(pathname: str, allow_list: str = VALID_PATH_CHARS,
               char_limit: int = CHAR_LIMIT, force_lower: bool = False) -> str:
    """
    Removes non-ASCII and other invalid path characters (on any reasonable
    OS) from a path, then trims to a maximum length.
    """
    
    return clean_filename(pathname, allow_list=allow_list, 
                          char_limit=char_limit, force_lower=force_lower)


def flatten_path(pathname: str, separator_chars: str = SEPARATOR_CHARS) -> str:
    """
    Removes non-ASCII and other invalid path characters (on any reasonable
    OS) from a path, then trims to a maximum length. Replaces all valid
    separators with '~'.
    """
    
    s = clean_path(pathname)
    for c in separator_chars:
        s = s.replace(c, '~')
    return s


#%% Platform-independent way to open files in their associated application

import sys,subprocess

def open_file(filename):
    if sys.platform == "win32":
        os.startfile(filename)
    else:
        opener = "open" if sys.platform == "darwin" else "xdg-open"
        subprocess.call([opener, filename])


#%% zipfile management functions

def unzip_file(input_file, output_folder=None):
    """
    Unzip a zipfile to the specified output folder, defaulting to the same location as
    the input file    
    """
    
    if output_folder is None:
        output_folder = os.path.dirname(input_file)
        
    with zipfile.ZipFile(input_file, 'r') as zf:
        zf.extractall(output_folder)


#%% File list functions

def write_list_to_file(output_file: str, strings: Sequence[str]) -> None:
    """
    Writes a list of strings to either a JSON file or text file,
    depending on extension of the given file name.
    """
    
    with open(output_file, 'w') as f:
        if output_file.endswith('.json'):
            json.dump(strings, f, indent=1)
        else:
            f.write('\n'.join(strings))


def read_list_from_file(filename: str) -> List[str]:
    """
    Reads a json-formatted list of strings from a file.
    """
    
    assert filename.endswith('.json')
    with open(filename, 'r') as f:
        file_list = json.load(f)
    assert isinstance(file_list, list)
    for s in file_list:
        assert isinstance(s, str)
    return file_list


#%% Zip functions

def zip_file(input_fn, output_fn=None, overwrite=False, verbose=False, compresslevel=9):
    """
    Zip a single file, by default writing to a new file called [input_fn].zip
    """
    
    basename = os.path.basename(input_fn)
    
    if output_fn is None:
        output_fn = input_fn + '.zip'
        
    if (not overwrite) and (os.path.isfile(output_fn)):
        print('Skipping existing file {}'.format(output_fn))
        return
    
    if verbose:
        print('Zipping {} to {}'.format(input_fn,output_fn))
    
    with ZipFile(output_fn,'w',zipfile.ZIP_DEFLATED) as zipf:
        zipf.write(input_fn,arcname=basename,compresslevel=compresslevel,
                   compress_type=zipfile.ZIP_DEFLATED)

    return output_fn


def zip_folder(input_folder, output_fn=None, overwrite=False, verbose=False, compresslevel=9):
    """
    Recursively zip everything in [input_folder], storing outputs as relative paths.
    
    Defaults to writing to [input_folder].zip
    """
    
    if output_fn is None:
        output_fn = input_folder + '.zip'
        
    if not overwrite:
        assert not os.path.isfile(output_fn), 'Zip file {} exists'.format(output_fn)
        
    if verbose:
        print('Zipping {} to {}'.format(input_folder,output_fn))
    
    relative_filenames = recursive_file_list(input_folder,return_relative_paths=True)
    
    with ZipFile(output_fn,'w',zipfile.ZIP_DEFLATED) as zipf:
        for input_fn_relative in relative_filenames:
            input_fn_abs = os.path.join(input_folder,input_fn_relative)            
            zipf.write(input_fn_abs,
                       arcname=input_fn_relative,
                       compresslevel=compresslevel,
                       compress_type=zipfile.ZIP_DEFLATED)

    return output_fn

        
def parallel_zip_files(input_files,max_workers=16):
    """
    Zip one or more files to separate output files in parallel, leaving the 
    original files in place.
    """

    n_workers = min(max_workers,len(input_files))
    pool = ThreadPool(n_workers)
    with tqdm(total=len(input_files)) as pbar:
        for i,_ in enumerate(pool.imap_unordered(zip_file,input_files)):
            pbar.update()



########
#
# matlab_porting_tools.py
#
# Module containing a few ported Matlab functions that made it easier
# for me to port other, larger Matlab functions.
#
########

#%% Constants and imports

import ntpath


#%% sec2hms()
        
def sec2hms(tSeconds):
    """
    function [str,h,m,s] = sec2hms(tSeconds,separator)
    
        Convert a time in seconds to a string of the form:
    
        1 hour, 2 minutes, 31.4 seconds
    
    I prefer using the humanfriendly package for this, but I use this when
    porting from Matlab.    
    """

    # https://stackoverflow.com/questions/775049/python-time-seconds-to-hms    
    m, s = divmod(tSeconds, 60)
    h, m = divmod(m, 60)
    
    # colonString = '%d:%02d:%02d' % (h, m, s)
    # return (colonString,verboseString)
    
    hms = ''
    separator = ', '
    if (h > 0):
        pluralString = ''    
        if (h > 1):
            pluralString = 's'
        hms = hms + '%d hour%s%s' % (h,pluralString,separator)
    
    if (m > 0):
        pluralString = ''
        if (m > 1):
            pluralString = 's'
        hms = hms + '%d min%s%s' % (m,pluralString,separator)
    
    hms = hms + '%3.3fsec' % s
            
    return hms

#%% Test driver for sec2hms()
    
if False:

    pass

    #%%
    
    TEST_VALUES = [60033, 30.4, 245234523454.1]
    
    for n in TEST_VALUES:
        s = sec2hms(n)
        print('{} - {}'.format(n,s))


#%% read_lines_from_file()

def read_lines_from_file(filename):
    
    with open(filename) as f:
        content = f.readlines()
    
    # Remove trailing newlines
    content = [x.rstrip() for x in content] 
    
    return content


#%% write_lines_to_file()
    
def write_lines_to_file(lines, filename):
    
    with open(filename,'w') as f:
        for line in lines:
            f.write(line+ '\n')
            

#%% string_ends_with()
            
def string_ends_with(s,query):    
    return s.endswith(query)

def string_starts_with(s,query):    
    return s.startswith(query)



########
#
# directory_listing.py
#
# Script for creating HTML directory listings for a local directory and
# all its subdirectories.  Primarily intended for use on mounted blob
# containers, so it includes the ability to set the content-type property
# on the generated html using the blob SDK, so it can be browser-viewable.
#
# Also includes a preview of a jpg file (the first in an alphabetical list),
# if present.
#
# Sample invocation:
#
# python directory_listing.py /naipout/v002 --basepath /naipout/v002 \
#   --enable_overwrite \
#   --sas_url "https://naipblobs.blob.core.windows.net/naip/v002?sv=..."
#
########

#%% Imports

import os
import sys
import argparse
import re

import azure.common
from azure.storage.blob import BlobServiceClient, ContentSettings


#%% Directory enumeration functions

def is_image_file(path):
    """
    Checks whether the provided file path points to an image by checking the
    file extension. The following file extensions are considered images: jpg,
    jpeg.  The check is case-insensitive.

    Args:
        path: string, path to image file or just the file name

    Returns:
        boolean, True if the file is an image
    """

    return os.path.splitext(path)[1].lower()[1:] in ['jpg', 'jpeg'] #, 'gif', 'tiff', 'tif', 'png']


def create_plain_index(root, dirs, files, dirname=None):
    """
    Creates the fairly plain HTML folder index
    including a preview of a single image file, if any is present.
    Returns the HTML source as string.

    Args:
        root: string, path to the root directory, all paths in *dirs* and
            *files* are relative to this one
        dirs: list of strings, the directories in *root*
        files: list of string, the files in *root*
        dirname: name to print in the html, which may be different than *root*

    Returns: HTML source of the directory listing
    """

    if dirname is None:
        dirname = root or '/'

    html = "<!DOCTYPE html>\n"
    html += "<html lang='en'><head>"
    html += "<title>Index of {}</title>\n".format(dirname)
    html += "<meta charset='UTF-8'>\n"
    html += "<style>\n"
    html += "body { font-family: Segoe UI, Helvetica, Arial, sans-serif; }\n"
    html += "a { text-decoration:none; }\n"
    html += "</style>\n"
    html += "</head><body>\n"

    html += "<h1>Index of {}</h1>\n".format(dirname)

    # Insert preview image
    jpg_files = [f for f in files if is_image_file(f)]

    if len(jpg_files) > 0:

        # This is slow, so defaulting to just the first image:
        #
        # Use the largest image file as this is most likely to contain
        # useful content.
        #
        # jpg_file_sizes = [os.path.getsize(f) for f in jpg_files]
        # largest_file_index = max(range(len(jpg_files)), key=lambda x: jpg_file_sizes[x])

        html += "<a href='{0}'><img style='height:200px; float:right;' src='{0}' alt='Preview image'></a>\n".format(jpg_files[0])
    else:
        html += "\n"
        # html += "<p style='width:15em; float:right; margin:0;'>[No preview available]</p>\n"

    if root:
        html += "<p><a href='../index.html'>To parent directory</a></p>\n"
    else:
        html += "\n"
        # html += "<p>This is the root directory.</p>\n"

    html += "<h2>Folders</h2>\n"
    if len(dirs) > 0:
        html += "<ul style='list-style-type: none; padding-left:1em;'>\n"
        for dir in sorted(dirs):
            html += "<li>&#128193; <a href='{0}/index.html'>{0}</a></li>\n".format(dir)
        html += "</ul>\n"
    else:
        html += "<p style='padding-left:1em;'>No folders</p>\n"

    html += "<h2>Files</h2>\n"
    if len(files) > 0:
        html += "<ul style='list-style-type: none; padding-left:1em;'>\n"
        for fname in sorted(files):
            if is_image_file(fname):
                html += "<li>&#128443; <a href='{0}'>{0}</a></li>\n".format(fname)
            else:
                html += "<li>&#128442; <a href='{0}'>{0}</a></li>\n".format(fname)
        html += "</ul>\n"
    else:
        html += "<p style = 'padding-left:1em;'>No files</p>\n"

    # Add some space at the bottom because the browser's status bar might hide stuff
    html += "<p style='margin:2em;'>&nbsp;</p>\n"
    html += "</body></html>\n"
    return html


def traverse_and_create_index(dir, sas_url=None, overwrite_files=False,
                              template_fun=create_plain_index, basepath=None):
    """
    Recursively traverses the local directory *dir* and generates a index
    file for each folder using *template_fun* to generate the HTML output.
    Excludes hidden files.

    Args:
        dir: string, path to directory
        template_fun: function taking three arguments (string, list of string, list of string)
            representing the current root, the list of folders, and the list of files.
            Should return the HTML source of the index file

    Return:
        None
    """

    print("Traversing {}".format(dir))

    # Make sure we remove the trailing /
    dir = os.path.normpath(dir)

    # If we want to set the content type in blob storage using a SAS URL
    if sas_url:

        # Example: sas_url = 'https://accname.blob.core.windows.net/bname/path/to/folder?st=...&se=...&sp=...&...'
        if '?' in sas_url:
            # 'https://accname.blob.core.windows.net/bname/path/to/folder' and 'st=...&se=...&sp=...&...'
            base_url, sas_token = sas_url.split('?', 1)
        else:
            # 'https://accname.blob.core.windows.net/bname/path/to/folder' and None
            base_url, sas_token = sas_url, None
        # Remove https:// from base url
        # 'accname.blob.core.windows.net/bname/path/to/folder'
        base_url = base_url.split("//", 1)[1]
        # Everything up to the first dot is account name
        # 'accname'
        account_name = base_url.split(".", 1)[0]
        # get everything after the first /
        # 'bname/path/to/folder'
        query_string = base_url.split("/", 1)[1]
        # Get container name and subfolder
        if '/' in query_string:
            # 'bname', 'path/to/folder'
            container_name, container_folder = query_string.split("/", 1)
        else:
            container_name, container_folder = query_string, ''

        # Prepare the storage access
        target_settings = ContentSettings(content_type='text/html')
        blob_service = BlobServiceClient(
            account_url=f'{account_name}.blob.core.windows.net',
            credential=sas_token)

    # Traverse directory and all sub directories, excluding hidden files
    for root, dirs, files in os.walk(dir):

        # Exclude files and folders that are hidden
        files = [f for f in files if not f[0] == '.']
        dirs[:] = [d for d in dirs if not d[0] == '.']

        # Output is written to file *root*/index.html
        output_file = os.path.join(root, "index.html")

        if not overwrite_files and os.path.isfile(output_file):
            print('Skipping {}, file exists'.format(output_file))
            continue

        print("Generating {}".format(output_file))

        # Generate HTML with template function
        dirname = None
        if basepath is not None:
            dirname = os.path.relpath(root,basepath)
        html = template_fun(root[len(dir):], dirs, files, dirname)

        # Write to file
        with open(output_file, 'wt') as fi:
            fi.write(html)

        # Set content type in blob storage
        if sas_url:
            if container_folder:
                output_blob_path = container_folder + '/' + output_file[len(dir) + 1:]
            else:
                output_blob_path = output_file[len(dir) + 1:]
            try:
                blob_client = blob_service.get_blob_client(container_name, output_blob_path)
                blob_client.set_http_headers(content_settings=target_settings)
            except azure.common.AzureMissingResourceHttpError:
                print('ERROR: It seems the SAS URL is incorrect or does not allow setting properties.')
                return


#%% Command-line driver

def parse_args(argv=sys.argv[1:]):

    parser = argparse.ArgumentParser()
    parser.add_argument("directory", type=str, help='Path to directory which should be traversed.')
    parser.add_argument("--basepath", type=str, help='Folder names will be printed relative to basepath, if specified', default=None)
    parser.add_argument("--sas_url", type=str, help='Blobfuse does not set the content-type property ' + \
        'properly and hence index.html won\'t be accessible in the browser. If you want to set the ' + \
        'content-type in the corresponding blob storage, provide the SAS URL that corresponds to the ' + \
        "directory, e.g. if *directory* is /mountpoint/path/to/folder, then *--sas_url* looks like " + \
        "'https://accname.blob.core.windows.net/bname/path/to/folder?st=...&se=...&sp=...&...'")
    parser.add_argument("--enable_overwrite", action='store_true', default=False,
                        help='If set, the script will overwrite existing index.html files.')
    return parser.parse_args(argv)


if __name__ == '__main__':

    args = parse_args()

    assert os.path.isdir(args.directory), "{} is not a valid directory".format(args.directory)
    assert re.match('https?://[^\.]+\.blob\.core\.windows\.net/.+', args.sas_url), "--sas_url does not " + \
        "match the format https://accname.blob.core.windows.net/bname/path/to/folder?..."

    traverse_and_create_index(args.directory, overwrite_files=args.enable_overwrite, sas_url=args.sas_url, basepath=args.basepath)



########
#
# process_utils.py
#
# Run something at the command line and capture the output, based on:
#
# https://stackoverflow.com/questions/4417546/constantly-print-subprocess-output-while-process-is-running
# 
# Includes handy example code for doing this on multiple processes/threads.
#
########

#%% Constants, imports, and environment

import os
import subprocess

os.environ["PYTHONUNBUFFERED"] = "1"

def execute(cmd):
  """
  Run [cmd] (a single string) in a shell, yielding each line of output to the caller.
  """
 
  # https://stackoverflow.com/questions/4417546/constantly-print-subprocess-output-while-process-is-running
  popen = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                           shell=True, universal_newlines=True)
  for stdout_line in iter(popen.stdout.readline, ""):
     yield stdout_line
  popen.stdout.close()
  return_code = popen.wait()
  if return_code:
    raise subprocess.CalledProcessError(return_code, cmd)


def execute_and_print(cmd,print_output=True):
  """
  Run [cmd] (a single string) in a shell, capturing and printing output.  Returns
  a dictionary with fields "status" and "output".
  """
 
  to_return = {'status':'unknown','output':''}
  output=[]
  try:
    for s in execute(cmd):
      output.append(s)
      if print_output:
        print(s,end='',flush=True)
    to_return['status'] = 0
  except subprocess.CalledProcessError as cpe:
    print('execute_and_print caught error: {}'.format(cpe.output))
    to_return['status'] = cpe.returncode
  to_return['output'] = output
   
  return to_return


#%% Single-threaded test driver for execute_and_print

if False:
    
    #%%
    
    if os.name == 'nt':
        execute_and_print('echo hello && ping -n 5 127.0.0.1 && echo goodbye')  
    else:
        execute_and_print('echo hello && sleep 1 && echo goodbye')  
 

#%% Parallel test driver for execute_and_print

if False:
   
    #%%
   
    from functools import partial
    from multiprocessing.pool import ThreadPool as ThreadPool
    from multiprocessing.pool import Pool as Pool
   
    n_workers = 10
   
    # Should we use threads (vs. processes) for parallelization?
    use_threads = True
   
    test_data = ['a','b','c','d']
   
    def process_sample(s):
        return execute_and_print('echo ' + s,True)
       
    if n_workers == 1:  
     
      results = []
      for i_sample,sample in enumerate(test_data):    
        results.append(process_sample(sample))
     
    else:
     
      n_threads = min(n_workers,len(test_data))
     
      if use_threads:
        print('Starting parallel thread pool with {} workers'.format(n_threads))
        pool = ThreadPool(n_threads)
      else:
        print('Starting parallel process pool with {} workers'.format(n_threads))
        pool = Pool(n_threads)
   
      results = list(pool.map(partial(process_sample),test_data))
      
      for r in results:
          print(r)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""
Helper functions for the batch processing API.
"""

import logging
import os
from datetime import datetime
from typing import Tuple, Any, Sequence, Optional

import sas_blob_utils


log = logging.getLogger(os.environ['FLASK_APP'])


#%% helper classes and functions

def make_error(error_code: int, error_message: str) -> Tuple[dict, int]:
    # TODO log exception when we have more telemetry
    log.error(f'Error {error_code} - {error_message}')
    return {'error': error_message}, error_code


def check_data_container_sas(input_container_sas: str) -> Optional[Tuple[int, str]]:
    """
    Returns a tuple (error_code, msg) if not a usable SAS URL, else returns None
    """
    # TODO check that the expiry date of input_container_sas is at least a month
    # into the future
    permissions = sas_blob_utils.get_permissions_from_uri(input_container_sas)
    data = sas_blob_utils.get_all_query_parts(input_container_sas)

    msg = ('input_container_sas provided does not have both read and list '
           'permissions.')
    if 'read' not in permissions or 'list' not in permissions:
        if 'si' in data:
            # if no permission specified explicitly but has an access policy, assumes okay
            # TODO - check based on access policy as well
            return None

        return 400, msg

    return None


def get_utc_time() -> str:
    # return current UTC time as a string in the ISO 8601 format (so we can query by
    # timestamp in the Cosmos DB job status table.
    # example: '2021-02-08T20:02:05.699689Z'
    return datetime.utcnow().isoformat(timespec='microseconds') + 'Z'


def get_job_status(request_status: str, message: Any) -> dict:
    return {
        'request_status': request_status,
        'message': message
    }


def validate_provided_image_paths(image_paths: Sequence[Any]) -> Tuple[Optional[str], bool]:
    """Given a list of image_paths (list length at least 1), validate them and
    determine if metadata is available.
    Args:
        image_paths: a list of string (image_id) or a list of 2-item lists
            ([image_id, image_metadata])
    Returns:
        error: None if checks passed, otherwise a string error message
        metadata_available: bool, True if available
    """
    # image_paths will have length at least 1, otherwise would have ended before this step
    first_item = image_paths[0]
    metadata_available = False
    if isinstance(first_item, str):
        for i in image_paths:
            if not isinstance(i, str):
                error = 'Not all items in image_paths are of type string.'
                return error, metadata_available
        return None, metadata_available
    elif isinstance(first_item, list):
        metadata_available = True
        for i in image_paths:
            if len(i) != 2:  # i should be [image_id, metadata_string]
                error = ('Items in image_paths are lists, but not all lists '
                         'are of length 2 [image locator, metadata].')
                return error, metadata_available
        return None, metadata_available
    else:
        error = 'image_paths contain items that are not strings nor lists.'
        return error, metadata_available



"""

map_classification_categories.py

Maps a classifier's output categories to desired target categories.

In this file, we use the following terminology:
* "category": a category output by the classifier
* "target": name of a desired group, comprising >= 1 classifier categories

Takes as input 2 label specification JSON files:

1) desired label specification JSON file
   this should not have a target named "other"
   
2) label specification JSON file of trained classifier

The mapping is accomplished as follows:
   
1. For each category in the classifier label spec, find all taxon nodes that
    belong to that category.

2. Given a target in the desired label spec, find all taxon nodes that belong
    to that target. If there is any classifier category whose nodes are a
    subset of the target nodes, then map the classifier category to that target.
    Any partial intersection between a target's nodes and a category's nodes
    is considered an error.

3. If there are any classifier categories that have not yet been assigned a
    target, group them into the "other" target.

This script outputs a JSON file that maps each target to a list of classifier
categories.

The taxonomy mapping parts of this script are very similar to json_validator.py.

"""

#%% Imports

from __future__ import annotations

import argparse
import json
import os

from collections import defaultdict
from collections.abc import Mapping
from typing import Any, Optional

import networkx as nx
import pandas as pd
from tqdm import tqdm

from taxonomy_mapping.taxonomy_graph import (
    build_taxonomy_graph, dag_to_tree, TaxonNode)


#%% Example usage

"""
    python map_classification_categories.py \
        desired_label_spec.json \
        /path/to/classifier/label_spec.json \
        $HOME/camera-traps-private/camera_trap_taxonomy_mapping.csv
"""


#%% Main function

def main(desired_label_spec_json_path: str,
         classifier_label_spec_json_path: str,
         taxonomy_csv_path: str,
         output_json_path: str,
         classifier_label_index_path: Optional[str]) -> None:

    print('Reading label spec JSON files')
    with open(desired_label_spec_json_path, 'r') as f:
        target_spec = json.load(f)
    with open(classifier_label_spec_json_path, 'r') as f:
        classifier_spec = json.load(f)

    if classifier_label_index_path is not None:
        with open(classifier_label_index_path, 'r') as f:
            classifier_labels = set(json.load(f).values())
        assert classifier_labels <= set(classifier_spec.keys())
        if len(classifier_labels) < len(classifier_spec):
            classifier_spec = {
                k: v for k, v in classifier_spec.items()
                if k in classifier_labels
            }

    print('Building taxonomy hierarchy')
    taxonomy_df = pd.read_csv(taxonomy_csv_path)
    graph, taxon_to_node, label_to_node = build_taxonomy_graph(taxonomy_df)
    dag_to_tree(graph, taxon_to_node)

    print('Mapping label spec to nodes')
    classifier_label_to_nodes = label_spec_to_nodes(
        classifier_spec, taxon_to_node, label_to_node)
    target_label_to_nodes = label_spec_to_nodes(
        target_spec, taxon_to_node, label_to_node)

    print('Creating mapping from target to classifier categories')
    target_to_classifier_labels = map_target_to_classifier(
        target_label_to_nodes, classifier_label_to_nodes)
    os.makedirs(os.path.dirname(output_json_path), exist_ok=True)
    with open(output_json_path, 'w') as f:
        json.dump(target_to_classifier_labels, f, indent=1)


#%% Support functions

def map_target_to_classifier(
        target_label_to_nodes: Mapping[str, set[TaxonNode]],
        classifier_label_to_nodes: Mapping[str, set[TaxonNode]]
        ) -> dict[str, list[str]]:
    """
    For each target, if there is any classifier category whose nodes are a
    subset of the target nodes, then assign the classifier category to that
    target. Any partial intersection between a target's nodes and a category's
    nodes is considered an error.

    Args:
        target_label_to_nodes: dict, maps target to set of nodes,
            all of the sets of nodes should be disjoint
        classifier_label_to_nodes: dict, maps classifier label to set of nodes,
            all of the sets of nodes should be disjoint

    Returns: dict, maps target label to set of classifier labels
    """
    
    remaining_classifier_labels = set(classifier_label_to_nodes.keys())
    target_to_classifier_labels: defaultdict[str, set[str]] = defaultdict(set)
    for target, target_nodes in tqdm(target_label_to_nodes.items()):
        for label, classifier_nodes in classifier_label_to_nodes.items():
            overlap = classifier_nodes & target_nodes
            if len(overlap) == len(classifier_nodes):
                target_to_classifier_labels[target].add(label)
                remaining_classifier_labels.remove(label)
            elif 0 < len(overlap) < len(classifier_nodes):  # partial overlap
                raise ValueError('Only partial overlap between target '
                                 f'{target} and classifier label {label}')
    if len(remaining_classifier_labels) > 0:
        target_to_classifier_labels['other'] = remaining_classifier_labels
    target_to_sorted_labels = {
        target: sorted(labels_set)
        for target, labels_set in target_to_classifier_labels.items()
    }
    return target_to_sorted_labels


def parse_spec(spec_dict: Mapping[str, Any],
               taxon_to_node: dict[tuple[str, str], TaxonNode],
               label_to_node: dict[tuple[str, str], TaxonNode]
               ) -> set[TaxonNode]:
    """
    Args:
        spec_dict: dict, contains keys ['taxa', 'dataset_labels', 'exclude']
            {
              "taxa": [
                {'level': 'family', 'name': 'cervidae', 'datasets': ['idfg']},
                {'level': 'genus',  'name': 'meleagris'} ],
              "dataset_labels": { "idfg_swwlf_2019": ["bird"] },
              "exclude": {...}
            }
        taxon_to_node: dict, maps (taxon_level, taxon_name) to a TaxonNode
        label_to_node: dict, maps (dataset_name, dataset_label) to the lowest
            TaxonNode node in the tree that contains the label

    Returns: set of TaxonNode, nodes selected by the taxa spec

    Raises: ValueError, if specification does not match any dataset labels
    """
    
    result = set()
    if 'taxa' in spec_dict:
        for taxon in spec_dict['taxa']:
            key = (taxon['level'].lower(), taxon['name'].lower())
            if key in taxon_to_node:
                node = taxon_to_node[key]
                result.add(node)
                result |= nx.descendants(node.graph, node)
            else:
                print(f'Taxon {key} not found in taxonomy graph. Ignoring.')
    if 'dataset_labels' in spec_dict:
        for ds, ds_labels in spec_dict['dataset_labels'].items():
            ds = ds.lower()
            for ds_label in ds_labels:
                node = label_to_node[(ds, ds_label.lower())]
                result.add(node)
                result |= nx.descendants(node.graph, node)
    if 'exclude' in spec_dict:
        result -= parse_spec(spec_dict['exclude'], taxon_to_node, label_to_node)
    if len(result) == 0:
        raise ValueError(f'specification matched no TaxonNode: {spec_dict}')
    return result


def label_spec_to_nodes(label_spec_js: dict[str, dict[str, Any]],
                        taxon_to_node: dict[tuple[str, str], TaxonNode],
                        label_to_node: dict[tuple[str, str], TaxonNode]
                        ) -> dict[str, set[TaxonNode]]:
    """
    Convert label spec to a mapping from classification labels to a set of
    nodes.

    Args:
        label_spec_js: dict, Python dict representation of JSON file
            see classification/README.md
        taxon_to_node: dict, maps (taxon_level, taxon_name) to a TaxonNode
        label_to_node: dict, maps (dataset_name, dataset_label) to the lowest
            TaxonNode node in the tree that contains the label

    Returns: dict, maps label name to set of TaxonNode

    Raises: ValueError, if a classification label specification matches no
        TaxonNode, or if a node is included in two or more classification labels
    """
    
    # maps output label name to set of (dataset, dataset_label) tuples
    seen_nodes: set[TaxonNode] = set()
    label_to_nodes: dict[str, set[TaxonNode]] = {}
    for label, spec_dict in label_spec_js.items():
        include_set = parse_spec(spec_dict, taxon_to_node, label_to_node)
        if include_set.isdisjoint(seen_nodes):
            label_to_nodes[label] = include_set
            seen_nodes |= include_set
        else:
            # find which other label (label_b) has intersection
            for label_b, set_b in label_to_nodes.items():
                shared = include_set.intersection(set_b)
                if len(shared) > 0:
                    print(f'Labels {label} and {label_b} share images:', shared)
                    raise ValueError('Intersection between sets!')
    return label_to_nodes


#%% Command-line driver

def _parse_args() -> argparse.Namespace:
    
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Create mapping from target categories to classifier '
                    'labels.')
    parser.add_argument(
        'desired_label_spec_json',
        help='path to JSON file containing desired label specification')
    parser.add_argument(
        'classifier_label_spec_json',
        help='path to JSON file containing label specification of a trained '
             'classifier')
    parser.add_argument(
        'taxonomy_csv',
        help='path to taxonomy CSV file')
    parser.add_argument(
        '-o', '--output', required=True,
        help='path to output JSON')
    parser.add_argument(
        '-i', '--classifier-label-index',
        help='(optional) path to label index JSON file for trained classifier, '
             'needed if not all labels from <classifier_label_spec_json> were '
             'actually used (e.g., if some labels were filtered out by the '
             '--min-locs argument for create_classification_dataset.py)')
    return parser.parse_args()


if __name__ == '__main__':
    
    args = _parse_args()
    main(desired_label_spec_json_path=args.desired_label_spec_json,
         classifier_label_spec_json_path=args.classifier_label_spec_json,
         taxonomy_csv_path=args.taxonomy_csv,
         output_json_path=args.output,
         classifier_label_index_path=args.classifier_label_index)



"""

coco_to_yolo.py

Converts a COCO-formatted dataset to a YOLO-formatted dataset, flattening
the dataset (to a single folder) in the process.

If the input and output folders are the same, writes .txt files to the input folder,
and neither moves nor modifies images.

Currently ignores segmentation masks, and errors if an annotation has a 
segmentation polygon but no bbox.

Has only been tested on a handful of COCO Camera Traps data sets; if you
use it for more general COCO conversion, YMMV.

"""

#%% Imports and constants

import json
import os
import shutil

from collections import defaultdict
from tqdm import tqdm

from md_utils.path_utils import safe_create_link,find_images


#%% Support functions

def write_yolo_dataset_file(yolo_dataset_file,
                            dataset_base_dir,
                            class_list,
                            train_folder_relative=None,
                            val_folder_relative=None,
                            test_folder_relative=None):
    """
    Write a YOLOv5 dataset.yaml file to the absolute path [yolo_dataset_file] (should
    have a .yaml extension, though it's only a warning if it doesn't).  
    
    Args:
        yolo_dataset_file (str): the file, typically ending in .yaml or .yml, to write.  
            Does not have to be within dataset_base_dir.
        dataset_base_dir (str): the absolute base path of the YOLO dataset
        class_list (list or str): an ordered list of class names (the first item will be class 0, 
            etc.), or the name of a text file containing an ordered list of class names (one per 
            line, starting from class zero).
    """
    
    # Read class names
    if isinstance(class_list,str):
        with open(class_list,'r') as f:
            class_lines = f.readlines()
        class_lines = [s.strip() for s in class_lines]    
        class_list = [s for s in class_lines if len(s) > 0]

    if not (yolo_dataset_file.endswith('.yml') or yolo_dataset_file.endswith('.yaml')):
        print('Warning: writing dataset file to a non-yml/yaml extension:\n{}'.format(
            yolo_dataset_file))
        
    # Write dataset.yaml
    with open(yolo_dataset_file,'w') as f:
        
        f.write('# Train/val sets\n')
        f.write('path: {}\n'.format(dataset_base_dir))
        if train_folder_relative is not None:
            f.write('train: {}\n'.format(train_folder_relative))
        if val_folder_relative is not None:
            f.write('val: {}\n'.format(val_folder_relative))
        if test_folder_relative is not None:
            f.write('val: {}\n'.format(test_folder_relative))
            
        f.write('\n')
        
        f.write('# Classes\n')
        f.write('names:\n')
        for i_class,class_name in enumerate(class_list):
            f.write('  {}: {}\n'.format(i_class,class_name))

# ...def write_yolo_dataset_file(...)

            
def coco_to_yolo(input_image_folder,
                 output_folder,
                 input_file,
                 source_format='coco',
                 overwrite_images=False,
                 create_image_and_label_folders=False,
                 class_file_name='classes.txt',
                 allow_empty_annotations=False,
                 clip_boxes=False,
                 image_id_to_output_image_json_file=None,
                 images_to_exclude=None,
                 path_replacement_char='#',
                 category_names_to_exclude=None,
                 category_names_to_include=None,
                 write_output=True,
                 flatten_paths=True):
    """
    Converts a COCO-formatted dataset to a YOLO-formatted dataset, optionally flattening the 
    dataset to a single folder in the process.
    
    If the input and output folders are the same, writes .txt files to the input folder,
    and neither moves nor modifies images.
    
    Currently ignores segmentation masks, and errors if an annotation has a 
    segmentation polygon but no bbox.
    
    Args:
        input_image_folder (str): the folder where images live; filenames in the COCO .json
            file [input_file] should be relative to this folder
        output_folder (str): the base folder for the YOLO dataset
        input_file (str): a .json file in COCO format; can be the same as [input_image_folder], in which case
            images are left alone.
        source_format (str, optional): can be 'coco' (default) or 'coco_camera_traps'.  The only difference
            is that when source_format is 'coco_camera_traps', we treat an image with a non-bbox
            annotation with a category id of 0 as a special case, i.e. that's how an empty image
            is indicated.  The original COCO standard is a little ambiguous on this issue.  If
            source_format is 'coco', we either treat images as empty or error, depending on the value
            of [allow_empty_annotations].  [allow_empty_annotations] has no effect if source_format is
            'coco_camera_traps'.
        create_image_and_label_folder (bool, optional): whether to create separate folders called 'images' and
            'labels' in the YOLO output folder.  If create_image_and_label_folders is False, 
            a/b/c/image001.jpg will become a#b#c#image001.jpg, and the corresponding text file will 
            be a#b#c#image001.txt.  If create_image_and_label_folders is True, a/b/c/image001.jpg will become 
            images/a#b#c#image001.jpg, and the corresponding text file will be 
            labels/a#b#c#image001.txt.    
        clip_boxes (bool, optional): whether to clip bounding box coordinates to the range [0,1] before
            converting to YOLO xywh format
        image_id_to_output_image_json_file (str, optional): an optional *output* file, to which we will write
            a mapping from image IDs to output file names
        images_to_exclude (list, optional): a list of image files (relative paths in the input folder) that we 
            should ignore
        path_replacement_char (str, optional): only relevant if [flatten_paths] is True; this is used to replace
            path separators, e.g. if [path_replacement_char] is '#' and [flatten_paths] is True, a/b/c/d.jpg
            becomes a#b#c#d.jpg
        category_names_to_exclude (str, optional): category names that should not be represented in the
            YOLO output; only impacts annotations, does not prevent copying images.  There's almost no reason
            you would want to specify this and [category_names_to_include]. 
        category_names_to_include (str, optional): allow-list of category names that should be represented in the
            YOLO output; only impacts annotations, does not prevent copying images.  There's almost no reason
            you would want to specify this and [category_names_to_exclude]. 
        write_output (bool, optional): determines whether we actually copy images and write annotations;
            setting this to False mostly puts this function in "dry run" "mode.  The class list
            file is written regardless of the value of write_output.
    
    Returns:
        dict: information about the coco --> yolo mapping, containing at least the fields:
        
        - class_list_filename: the filename to which we wrote the flat list of class names required 
          by the YOLO format.
        - source_image_to_dest_image: a dict mapping source images to destination images
        - coco_id_to_yolo_id: a dict mapping COCO category IDs to YOLO category IDs        
    """
        
    ## Validate input
    
    if category_names_to_include is not None and category_names_to_exclude is not None:
        raise ValueError('category_names_to_include and category_names_to_exclude are mutually exclusive')
        
    if output_folder is None:
        output_folder = input_image_folder
    
    if images_to_exclude is not None:
        images_to_exclude = set(images_to_exclude)
            
    if category_names_to_exclude is None:
        category_names_to_exclude = {}
            
    assert os.path.isdir(input_image_folder)
    assert os.path.isfile(input_file)
    os.makedirs(output_folder,exist_ok=True)
    
    if (output_folder == input_image_folder) and (overwrite_images) and \
        (not create_image_and_label_folders) and (not flatten_paths):
            print('Warning: output folder and input folder are the same, disabling overwrite_images')
            overwrite_images = False
            
    ## Read input data
    
    with open(input_file,'r') as f:
        data = json.load(f)
        
        
    ## Parse annotations
  
    image_id_to_annotations = defaultdict(list)
    
    # i_ann = 0; ann = data['annotations'][0]
    for i_ann,ann in enumerate(data['annotations']):
        
        # Make sure no annotations have *only* segmentation data 
        if ( \
            ('segmentation' in ann.keys()) and \
            (ann['segmentation'] is not None) and \
            (len(ann['segmentation']) > 0) ) \
            and \
            (('bbox' not in ann.keys()) or (ann['bbox'] is None) or (len(ann['bbox'])==0)):
                raise ValueError('Oops: segmentation data present without bbox information, ' + \
                                 'this script isn\'t ready for this dataset')
        
        image_id_to_annotations[ann['image_id']].append(ann)
        
    print('Parsed annotations for {} images'.format(len(image_id_to_annotations)))
        
    # Re-map class IDs to make sure they run from 0...n-classes-1
    #
    # Note: this allows unused categories in the output data set.  This is OK for
    # some training pipelines, not for others.
    next_category_id = 0
    coco_id_to_yolo_id = {}    
    coco_id_to_name = {}
    yolo_id_to_name = {}
    coco_category_ids_to_exclude = set()
    
    for category in data['categories']:
        coco_id_to_name[category['id']] = category['name']
        if (category_names_to_include is not None) and \
            (category['name'] not in category_names_to_include):
            coco_category_ids_to_exclude.add(category['id'])
            continue
        elif (category['name'] in category_names_to_exclude):
            coco_category_ids_to_exclude.add(category['id'])
            continue               
        assert category['id'] not in coco_id_to_yolo_id
        coco_id_to_yolo_id[category['id']] = next_category_id
        yolo_id_to_name[next_category_id] = category['name']
        next_category_id += 1
        
    
    ## Process images (everything but I/O)
    
    # List of dictionaries with keys 'source_image','dest_image','bboxes','dest_txt'
    images_to_copy = []
    
    missing_images = []
    excluded_images = []
    
    image_names = set()
    
    typical_image_extensions = set(['.jpg','.jpeg','.png','.gif','.tif','.bmp'])
    
    printed_empty_annotation_warning = False
    
    image_id_to_output_image_name = {}
    
    print('Processing annotations')
    
    n_clipped_boxes = 0
    n_total_boxes = 0
    
    # i_image = 0; im = data['images'][i_image]
    for i_image,im in tqdm(enumerate(data['images']),total=len(data['images'])):
                
        output_info = {}
        source_image = os.path.join(input_image_folder,im['file_name'])        
        output_info['source_image'] = source_image
        
        if images_to_exclude is not None and im['file_name'] in images_to_exclude:
            excluded_images.append(im['file_name'])
            continue
        
        tokens = os.path.splitext(im['file_name'])
        if tokens[1].lower() not in typical_image_extensions:
            print('Warning: unusual image file name {}'.format(im['file_name']))
                
        if flatten_paths:
            image_name = tokens[0].replace('\\','/').replace('/',path_replacement_char) + \
                '_' + str(i_image).zfill(6)            
        else:
            image_name = tokens[0]

        assert image_name not in image_names, 'Image name collision for {}'.format(image_name)
        image_names.add(image_name)
        
        assert im['id'] not in image_id_to_output_image_name
        image_id_to_output_image_name[im['id']] = image_name
        
        dest_image_relative = image_name + tokens[1]
        output_info['dest_image_relative'] = dest_image_relative
        dest_txt_relative = image_name + '.txt'
        output_info['dest_txt_relative'] = dest_txt_relative
        output_info['bboxes'] = []
        
        # assert os.path.isfile(source_image), 'Could not find image {}'.format(source_image)
        if not os.path.isfile(source_image):
            print('Warning: could not find image {}'.format(source_image))
            missing_images.append(im['file_name'])
            continue
        
        image_id = im['id']
        
        image_bboxes = []
            
        if image_id in image_id_to_annotations:
                        
            for ann in image_id_to_annotations[image_id]:
                
                # If this annotation has no bounding boxes...
                if 'bbox' not in ann or ann['bbox'] is None or len(ann['bbox']) == 0:
                    
                    if source_format == 'coco':
                    
                        if not allow_empty_annotations:
                            # This is not entirely clear from the COCO spec, but it seems to be consensus
                            # that if you want to specify an image with no objects, you don't include any
                            # annotations for that image.
                            raise ValueError('If an annotation exists, it should have content')
                        else:
                            continue
                    
                    elif source_format == 'coco_camera_traps':
                        
                        # We allow empty bbox lists in COCO camera traps; this is typically a negative
                        # example in a dataset that has bounding boxes, and 0 is typically the empty 
                        # category.
                        if ann['category_id'] != 0:
                            if not printed_empty_annotation_warning:
                                printed_empty_annotation_warning = True
                                print('Warning: non-bbox annotation found with category {}'.format(
                                    ann['category_id']))
                        continue
                    
                    else:
                        
                        raise ValueError('Unrecognized COCO variant: {}'.format(source_format))
                    
                # ...if this is an empty annotation
                
                coco_bbox = ann['bbox']
                
                # This category isn't in our category list.  This typically corresponds to whole sets
                # of images that were excluded from the YOLO set.
                if ann['category_id'] in coco_category_ids_to_exclude:
                    continue
                
                yolo_category_id = coco_id_to_yolo_id[ann['category_id']]
                
                # COCO: [x_min, y_min, width, height] in absolute coordinates
                # YOLO: [class, x_center, y_center, width, height] in normalized coordinates
                
                # Convert from COCO coordinates to YOLO coordinates
                img_w = im['width']
                img_h = im['height']
                                
                if source_format in ('coco','coco_camera_traps'):
                    
                    x_min_absolute = coco_bbox[0]
                    y_min_absolute = coco_bbox[1]
                    box_w_absolute = coco_bbox[2]
                    box_h_absolute = coco_bbox[3]
                    
                    x_center_absolute = (x_min_absolute + (x_min_absolute + box_w_absolute)) / 2
                    y_center_absolute = (y_min_absolute + (y_min_absolute + box_h_absolute)) / 2
                    
                    x_center_relative = x_center_absolute / img_w
                    y_center_relative = y_center_absolute / img_h
                    
                    box_w_relative = box_w_absolute / img_w
                    box_h_relative = box_h_absolute / img_h
                    
                else:
                    
                    raise ValueError('Unrecognized source format {}'.format(source_format))
                
                if clip_boxes:
                    
                    clipped_box = False
                    
                    box_right = x_center_relative + (box_w_relative / 2.0)                    
                    if box_right > 1.0:
                        clipped_box = True
                        overhang = box_right - 1.0
                        box_w_relative -= overhang
                        x_center_relative -= (overhang / 2.0)

                    box_bottom = y_center_relative + (box_h_relative / 2.0)                                        
                    if box_bottom > 1.0:
                        clipped_box = True
                        overhang = box_bottom - 1.0
                        box_h_relative -= overhang
                        y_center_relative -= (overhang / 2.0)
                    
                    box_left = x_center_relative - (box_w_relative / 2.0)
                    if box_left < 0.0:
                        clipped_box = True
                        overhang = abs(box_left)
                        box_w_relative -= overhang
                        x_center_relative += (overhang / 2.0)
                        
                    box_top = y_center_relative - (box_h_relative / 2.0)
                    if box_top < 0.0:
                        clipped_box = True
                        overhang = abs(box_top)
                        box_h_relative -= overhang
                        y_center_relative += (overhang / 2.0)
                        
                    if clipped_box:
                        n_clipped_boxes += 1
                
                yolo_box = [yolo_category_id,
                            x_center_relative, y_center_relative, 
                            box_w_relative, box_h_relative]
                
                image_bboxes.append(yolo_box)
                n_total_boxes += 1
                
            # ...for each annotation 
            
        # ...if this image has annotations
        
        output_info['bboxes'] = image_bboxes
        
        images_to_copy.append(output_info)        
    
    # ...for each image
        
    print('\nWriting {} boxes ({} clipped) for {} images'.format(n_total_boxes,
                                                               n_clipped_boxes,len(images_to_copy)))
    print('{} missing images (of {})'.format(len(missing_images),len(data['images'])))
    
    if images_to_exclude is not None:
        print('{} excluded images (of {})'.format(len(excluded_images),len(data['images'])))
            
    
    ## Write output
    
    print('Generating class list')
    
    class_list_filename = os.path.join(output_folder,class_file_name)
    with open(class_list_filename, 'w') as f:
        print('Writing class list to {}'.format(class_list_filename))
        for i_class in range(0,len(yolo_id_to_name)):
            # Category IDs should range from 0..N-1
            assert i_class in yolo_id_to_name
            f.write(yolo_id_to_name[i_class] + '\n')
    
    if image_id_to_output_image_json_file is not None:
        print('Writing image ID mapping to {}'.format(image_id_to_output_image_json_file))
        with open(image_id_to_output_image_json_file,'w') as f:
            json.dump(image_id_to_output_image_name,f,indent=1)
    

    if (output_folder == input_image_folder) and (not create_image_and_label_folders):
        print('Creating annotation files (not copying images, input and output folder are the same)')
    else:
        print('Copying images and creating annotation files')

    if create_image_and_label_folders:
        dest_image_folder = os.path.join(output_folder,'images')
        dest_txt_folder = os.path.join(output_folder,'labels')
    else:
        dest_image_folder = output_folder
        dest_txt_folder = output_folder
        
    source_image_to_dest_image = {}
    
    # TODO: parallelize this loop
    #
    # output_info = images_to_copy[0]
    for output_info in tqdm(images_to_copy):

        source_image = output_info['source_image']
        dest_image_relative = output_info['dest_image_relative']
        dest_txt_relative = output_info['dest_txt_relative']
        
        dest_image = os.path.join(dest_image_folder,dest_image_relative)
        dest_txt = os.path.join(dest_txt_folder,dest_txt_relative)
        
        source_image_to_dest_image[source_image] = dest_image
        
        if write_output:
            
            os.makedirs(os.path.dirname(dest_image),exist_ok=True)        
            os.makedirs(os.path.dirname(dest_txt),exist_ok=True)
            
            if not create_image_and_label_folders:
                assert os.path.dirname(dest_image) == os.path.dirname(dest_txt)
            
            if (not os.path.isfile(dest_image)) or (overwrite_images):
                shutil.copyfile(source_image,dest_image)
        
            bboxes = output_info['bboxes']        
            
            # Only write an annotation file if there are bounding boxes.  Images with 
            # no .txt files are treated as hard negatives, at least by YOLOv5:
            #
            # https://github.com/ultralytics/yolov5/issues/3218
            #
            # I think this is also true for images with empty .txt files, but 
            # I'm using the convention suggested on that issue, i.e. hard 
            # negatives are expressed as images without .txt files.
            if len(bboxes) > 0:
                
                with open(dest_txt,'w') as f:
                    
                    # bbox = bboxes[0]
                    for bbox in bboxes:
                        assert len(bbox) == 5
                        s = '{} {} {} {} {}'.format(bbox[0],bbox[1],bbox[2],bbox[3],bbox[4])
                        f.write(s + '\n')
                        
        # ...if we're actually writing output
        
    # ...for each image
    
    coco_to_yolo_info = {}
    coco_to_yolo_info['class_list_filename'] = class_list_filename
    coco_to_yolo_info['source_image_to_dest_image'] = source_image_to_dest_image
    coco_to_yolo_info['coco_id_to_yolo_id'] = coco_id_to_yolo_id
    
    return coco_to_yolo_info

# ...def coco_to_yolo(...)


def create_yolo_symlinks(source_folder,images_folder,labels_folder,
                         class_list_file=None,
                         class_list_output_name='object.data',
                         force_lowercase_image_extension=False):
    """
    Given a YOLO-formatted folder of images and .txt files, creates a folder
    of symlinks to all the images, and a folder of symlinks to all the labels. 
    Used to support preview/editing tools that assume images and labels are in separate 
    folders.
    
    :meta private:
    """    
    
    assert source_folder != images_folder and source_folder != labels_folder
    
    os.makedirs(images_folder,exist_ok=True)
    os.makedirs(labels_folder,exist_ok=True)
    
    image_files_relative = find_images(source_folder,recursive=True,return_relative_paths=True)
    
    # image_fn_relative = image_files_relative[0]=    
    for image_fn_relative in tqdm(image_files_relative):
        
        source_file_abs = os.path.join(source_folder,image_fn_relative)
        target_file_abs = os.path.join(images_folder,image_fn_relative)
        
        if force_lowercase_image_extension:
            tokens = os.path.splitext(target_file_abs)
            target_file_abs = tokens[0] + tokens[1].lower()
            
        os.makedirs(os.path.dirname(target_file_abs),exist_ok=True)
        safe_create_link(source_file_abs,target_file_abs)
        source_annotation_file_abs = os.path.splitext(source_file_abs)[0] + '.txt'
        if os.path.isfile(source_annotation_file_abs):
            target_annotation_file_abs = \
                os.path.splitext(os.path.join(labels_folder,image_fn_relative))[0] + '.txt'
            os.makedirs(os.path.dirname(target_annotation_file_abs),exist_ok=True)
            safe_create_link(source_annotation_file_abs,target_annotation_file_abs)

    # ...for each image  

    if class_list_file is not None:
        target_class_list_file = os.path.join(labels_folder,class_list_output_name)
        safe_create_link(class_list_file,target_class_list_file)

# ...def create_yolo_symlinks(...)


#%% Interactive driver

if False:
    
    pass

    #%% Options
    
    input_file = os.path.expanduser('~/data/md-test-coco.json')
    image_folder = os.path.expanduser('~/data/md-test')    
    output_folder = os.path.expanduser('~/data/md-test-yolo')    
    create_image_and_label_folders=False
    class_file_name='classes.txt'
    allow_empty_annotations=False
    clip_boxes=False
    image_id_to_output_image_json_file=None
    images_to_exclude=None
    path_replacement_char='#'
    category_names_to_exclude=None
                                  
                                               
    #%% Programmatic execution
    
    coco_to_yolo_results = coco_to_yolo(image_folder,output_folder,input_file,
                     source_format='coco',
                     overwrite_images=False,
                     create_image_and_label_folders=create_image_and_label_folders,
                     class_file_name=class_file_name,
                     allow_empty_annotations=allow_empty_annotations,
                     clip_boxes=clip_boxes)                     
    
    create_yolo_symlinks(source_folder=output_folder,
                         images_folder=output_folder + '/images',
                         labels_folder=output_folder + '/labels',
                         class_list_file=coco_to_yolo_results['class_list_filename'],
                         class_list_output_name='object.data',
                         force_lowercase_image_extension=True)


    #%% Prepare command-line example

    s = 'python coco_to_yolo.py {} {} {} --create_bounding_box_editor_symlinks'.format(
        image_folder,output_folder,input_file)
    print(s)
    import clipboard; clipboard.copy(s)    


#%% Command-line driver

import sys,argparse

def main():

    parser = argparse.ArgumentParser(
        description='Convert COCO-formatted data to YOLO format, flattening the image structure')
    
    # input_image_folder,output_folder,input_file
    
    parser.add_argument(
        'input_folder',
        type=str,
        help='Path to input images')
    
    parser.add_argument(
        'output_folder',
        type=str,
        help='Path to flat, YOLO-formatted dataset')
    
    parser.add_argument(
        'input_file',
        type=str,
        help='Path to COCO dataset file (.json)')
    
    parser.add_argument(
        '--create_bounding_box_editor_symlinks',
        action='store_true',
        help='Prepare symlinks so the whole folder appears to contain "images" and "labels" folderss')        
    
    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    coco_to_yolo_results = coco_to_yolo(args.input_folder,args.output_folder,args.input_file)
    
    if args.create_bounding_box_editor_symlinks:
        create_yolo_symlinks(source_folder=args.output_folder,
                             images_folder=args.output_folder + '/images',
                             labels_folder=args.output_folder + '/labels',
                             class_list_file=coco_to_yolo_results['class_list_filename'],
                             class_list_output_name='object.data',
                             force_lowercase_image_extension=True)
    
if __name__ == '__main__':
    main()



"""

url_utils.py

Frequently-used functions for downloading or manipulating URLs

"""

#%% Imports and constants

import os
import re
import urllib
import tempfile
import requests

from functools import partial
from tqdm import tqdm
from urllib.parse import urlparse
from multiprocessing.pool import ThreadPool
from multiprocessing.pool import Pool

url_utils_temp_dir = None
max_path_len = 255


#%% Download functions

class DownloadProgressBar():
    """
    Progress updater based on the progressbar2 package.
    
    https://stackoverflow.com/questions/37748105/how-to-use-progressbar-module-with-urlretrieve
    """
    
    def __init__(self):
        self.pbar = None

    def __call__(self, block_num, block_size, total_size):
        if not self.pbar:
            # This is a pretty random import I'd rather not depend on outside of the
            # rare case where it's used, so importing locally
            # pip install progressbar2
            import progressbar            
            self.pbar = progressbar.ProgressBar(max_value=total_size)
            self.pbar.start()
            
        downloaded = block_num * block_size
        if downloaded < total_size:
            self.pbar.update(downloaded)
        else:
            self.pbar.finish()
            

def get_temp_folder(preferred_name='url_utils'):
    """
    Gets a temporary folder for use within this module.
    
    Args:
        preferred_name (str, optional): subfolder to use within the system temp folder
        
    Returns:
        str: the full path to the temporary subfolder
    """
    global url_utils_temp_dir
    
    if url_utils_temp_dir is None:
        url_utils_temp_dir = os.path.join(tempfile.gettempdir(),preferred_name)
        os.makedirs(url_utils_temp_dir,exist_ok=True)
        
    return url_utils_temp_dir
    
           
def download_url(url, 
                 destination_filename=None, 
                 progress_updater=None, 
                 force_download=False, 
                 verbose=True):
    """
    Downloads a URL to a file.  If no file is specified, creates a temporary file, 
    making a best effort to avoid filename collisions.
    
    Prints some diagnostic information and makes sure to omit SAS tokens from printouts.
    
    Args:
        url (str): the URL to download
        destination_filename (str, optional): the target filename; if None, will create
            a file in system temp space        
        progress_updater (object or bool, optional): can be "None", "False", "True", or a 
            specific callable object.  If None or False, no progress updated will be 
            displayed.  If True, a default progress bar will be created.
        force_download (bool, optional): download this file even if [destination_filename]
            exists.
        verbose (bool, optional): enable additional debug console output
        
    Returns:
        str: the filename to which [url] was downloaded, the same as [destination_filename]
        if [destination_filename] was not None
    """
    
    if progress_updater is not None and isinstance(progress_updater,bool):
        if not progress_updater:
            progress_updater = None
        else:
            progress_updater = DownloadProgressBar()
            
    url_no_sas = url.split('?')[0]
        
    if destination_filename is None:
        target_folder = get_temp_folder()
        url_without_sas = url.split('?', 1)[0]
        
        # This does not guarantee uniqueness, hence "semi-best-effort"
        url_as_filename = re.sub(r'\W+', '', url_without_sas)
        n_folder_chars = len(url_utils_temp_dir)
        if len(url_as_filename) + n_folder_chars > max_path_len:
            print('Warning: truncating filename target to {} characters'.format(max_path_len))
            url_as_filename = url_as_filename[-1*(max_path_len-n_folder_chars):]
        destination_filename = \
            os.path.join(target_folder,url_as_filename)
        
    if (not force_download) and (os.path.isfile(destination_filename)):
        if verbose:
            print('Bypassing download of already-downloaded file {}'.format(os.path.basename(url_no_sas)))
    else:
        if verbose:
            print('Downloading file {} to {}'.format(os.path.basename(url_no_sas),destination_filename),end='')
        target_dir = os.path.dirname(destination_filename)
        os.makedirs(target_dir,exist_ok=True)
        urllib.request.urlretrieve(url, destination_filename, progress_updater)  
        assert(os.path.isfile(destination_filename))
        nBytes = os.path.getsize(destination_filename)
        if verbose:
            print('...done, {} bytes.'.format(nBytes))
        
    return destination_filename


def download_relative_filename(url, output_base, verbose=False):
    """
    Download a URL to output_base, preserving relative path.  Path is relative to 
    the site, so:
        
        https://abc.com/xyz/123.txt
    
    ...will get downloaded to:
        
        output_base/xyz/123.txt        
        
    Args:
        url (str): the URL to download
        output_base (str): the base folder to which we should download this file
        verbose (bool, optional): enable additional debug console output
        
    Returns:
        str: the local destination filename
    """
    
    p = urlparse(url)
    # remove the leading '/'
    assert p.path.startswith('/'); relative_filename = p.path[1:]
    destination_filename = os.path.join(output_base,relative_filename)
    return download_url(url, destination_filename, verbose=verbose)


def _do_parallelized_download(download_info,overwrite=False,verbose=False):
    """
    Internal function for download parallelization.
    """
    
    url = download_info['url']
    target_file = download_info['target_file']
    result = {'status':'unknown','url':url,'target_file':target_file}
    
    if ((os.path.isfile(target_file)) and (not overwrite)):
        if verbose:
            print('Skipping existing file {}'.format(target_file))
        result['status'] = 'skipped'
        return result
    try:
        download_url(url=url, 
                     destination_filename=target_file,
                     verbose=verbose, 
                     force_download=overwrite)
    except Exception as e:
        print('Warning: error downloading URL {}: {}'.format(
            url,str(e)))     
        result['status'] = 'error: {}'.format(str(e))
        return result
    
    result['status'] = 'success'
    return result


def parallel_download_urls(url_to_target_file,verbose=False,overwrite=False,
                           n_workers=20,pool_type='thread'):
    """
    Downloads a list of URLs to local files.
    
    Catches exceptions and reports them in the returned "results" array.    
    
    Args:
        url_to_target_file: a dict mapping URLs to local filenames.
        verbose (bool, optional): enable additional debug console output
        overwrite (bool, optional): whether to overwrite existing local files
        n_workers (int, optional): number of concurrent workers, set to <=1 to disable
            parallelization
        pool_type (str, optional): worker type to use; should be 'thread' or 'process'
        
    Returns:
        list: list of dicts with keys:
            - 'url': the url this item refers to
            - 'status': 'skipped', 'success', or a string starting with 'error'
            - 'target_file': the local filename to which we downloaded (or tried to 
              download) this URL            
    """
    
    all_download_info = []
    
    print('Preparing download list')
    for url in tqdm(url_to_target_file):
        download_info = {}
        download_info['url'] = url
        download_info['target_file'] = url_to_target_file[url]
        all_download_info.append(download_info)
        
    print('Downloading {} images on {} workers'.format(
        len(all_download_info),n_workers))

    if n_workers <= 1:

        results = []
        
        for download_info in tqdm(all_download_info):        
            result = _do_parallelized_download(download_info,overwrite=overwrite,verbose=verbose)
            results.append(result)
        
    else:

        if pool_type == 'thread':
            pool = ThreadPool(n_workers)
        else:
            assert pool_type == 'process', 'Unsupported pool type {}'.format(pool_type)
            pool = Pool(n_workers)
        
        print('Starting a {} pool with {} workers'.format(pool_type,n_workers))
        
        results = list(tqdm(pool.imap(
            partial(_do_parallelized_download,overwrite=overwrite,verbose=verbose),
            all_download_info), total=len(all_download_info)))
                
    return results

    
def test_url(url, error_on_failure=True, timeout=None):
    """
    Tests the availability of [url], returning an http status code.
    
    Args:
        url (str): URL to test
        error_on_failure (bool, optional): whether to error (vs. just returning an
            error code) if accessing this URL fails
        timeout (int, optional): timeout in seconds to wait before considering this 
            access attempt to be a failure; see requests.head() for precise documentation
    
    Returns:
        int: http status code (200 for success)
    """
    
    # r = requests.get(url, stream=True, verify=True, timeout=timeout)
    r = requests.head(url, stream=True, verify=True, timeout=timeout)
    
    if error_on_failure and r.status_code != 200:        
        raise ValueError('Could not access {}: error {}'.format(url,r.status_code))
    return r.status_code
    

def test_urls(urls, error_on_failure=True, n_workers=1, pool_type='thread', timeout=None):
    """
    Verify that URLs are available (i.e., returns status 200).  By default,
    errors if any URL is unavailable.  
    
    Args:
        urls (list): list of URLs to test
        error_on_failure (bool, optional): whether to error (vs. just returning an
            error code) if accessing this URL fails
        n_workers (int, optional): number of concurrent workers, set to <=1 to disable
            parallelization
        pool_type (str, optional): worker type to use; should be 'thread' or 'process'
        timeout (int, optional): timeout in seconds to wait before considering this 
            access attempt to be a failure; see requests.head() for precise documentation
    
    Returns:
        list: a list of http status codes, the same length and order as [urls]
    """
        
    if n_workers <= 1:

        status_codes = []
        
        for url in tqdm(urls):
            
            r = requests.get(url, timeout=timeout)
            
            if error_on_failure and r.status_code != 200:        
                raise ValueError('Could not access {}: error {}'.format(url,r.status_code))
            status_codes.append(r.status_code)
                
    else:

        if pool_type == 'thread':
            pool = ThreadPool(n_workers)
        else:
            assert pool_type == 'process', 'Unsupported pool type {}'.format(pool_type)
            pool = Pool(n_workers)
        
        print('Starting a {} pool with {} workers'.format(pool_type,n_workers))
        
        status_codes = list(tqdm(pool.imap(
            partial(test_url,error_on_failure=error_on_failure,timeout=timeout),
            urls), total=len(urls)))
                
    return status_codes



#
# make_eMammal_json.py
#
# Produces the COCO-formatted json database for an eMammal dataset, i.e. a 
# collection of folders, each of which contains a deployment_manifest.xml file.
#
# In this process, each image needs to be loaded to size it.
#
# To add bounding box annotations to the resulting database, use 
# add_annotations_to_eMammal_json.py.
#

#%% Constants and imports

# Either add the eMammal directory to your path, or run from there
# os.chdir(r'd:\git\MegaDetector\database_tools\eMammal')

import json
import multiprocessing
import os
# import warnings
import eMammal_helpers as helpers

from datetime import datetime, date
from multiprocessing.dummy import Pool as ThreadPool
from lxml import etree
from tqdm import tqdm

# ignoring all "PIL cannot read EXIF metainfo for the images" warnings
# warnings.filterwarnings('ignore')

# Should we run the image size retrieval in parallel?
run_parallel = False

output_dir_path = r'd:\path'
deployments_path = r'd:\other_path'
db_filename = 'apr.json'
corrupt_images_db_filename = 'apr_corrupt.json'

description = 'description'
version = '1.0'
contributor = 'contributor'
curator = '.json created by curator'


#%% Support functions

def _add_image(entry, full_img_path):    
    """ 
    Open the image to get size information and add height and width to the image entry. 
    """
    
    img_width, img_height = helpers.get_img_size(full_img_path)
    if img_width == -1 or img_height == -1:
        corrupt_images.append(full_img_path)
        return None
    entry['width'] = img_width
    entry['height'] = img_height
    pbar.update(1)
    return entry


#%% Main loop (metadata processing; image sizes are retrieved later)
    
print('Creating tasks to get all images...')
start_time = datetime.now()
tasks = []
folders = os.listdir(deployments_path)

all_species_strings = set()

# deployment = folders[0]
for deployment in tqdm(folders):
    
    deployment_path = os.path.join(deployments_path, deployment)
    manifest_path = os.path.join(deployment_path, 'deployment_manifest.xml')

    assert os.path.isfile(manifest_path)
    
    with open(manifest_path, 'r') as f:
        tree = etree.parse(f)

    root = tree.getroot()
    project_id = root.findtext('ProjectId')
    deployment_id = root.findtext('CameraDeploymentID')
    deployment_location = root.findtext('CameraSiteName')

    image_sequences = root.findall('ImageSequence')
    
    # sequence = image_sequences[0]
    for sequence in image_sequences:
        
        seq_id = sequence.findtext('ImageSequenceId')

        # get species info for this sequence
        researcher_identifications = sequence.findall('ResearcherIdentifications')
        species = set()

        for researcher_id in researcher_identifications:
            identifications = researcher_id.findall('Identification')
            for id in identifications:
                species_common_name = helpers.clean_species_name(id.findtext('SpeciesCommonName'))
                species.add(species_common_name)

        species_str = ';'.join(sorted(list(species)))
        all_species_strings.add(species_str)
        
        # add each image's info to database
        images = sequence.findall('Image')
        seq_num_frames = len(images) # total number of frames in this sequence
        assert isinstance(seq_num_frames,int) and seq_num_frames > 0 # historical quirk
        
        # img = images[0]
        for img in images:
            
            img_id = img.findtext('ImageId')
            img_file_name = img.findtext('ImageFileName')
            assert img_file_name.lower().endswith('.jpg')

            img_frame = img.findtext('ImageOrder')
            if img_frame == '' or img_frame is None:
                # some manifests don't have the ImageOrder info, but the info is in the file name
                img_frame = img_file_name.split('i')[1].split('.')[0]

            # full_img_id has no frame info
            #
            # frame number only used in requests to iMerit for ordering
            full_img_id = 'datasetemammal.project{}.deployment{}.seq{}.img{}'.format(
                project_id, deployment_id, seq_id, img_id)            
            full_img_path = os.path.join(deployment_path, img_file_name)

            img_datetime, datetime_err = helpers.parse_timestamp(img.findtext('ImageDateTime'))
            if datetime_err:
                print('WARNING datetime parsing error for image {}. Error: {}'.format(
                    full_img_path, datetime_err))

            entry = {
                'id': full_img_id,
                'width': 0,  # place holders
                'height': 0,
                'file_name': os.path.join(deployment, img_file_name),
                'location': deployment_location,
                'datetime': str(img_datetime),
                'seq_id': seq_id,
                'frame_num': int(img_frame),
                'seq_num_frames': seq_num_frames,
                'label': species_str  # extra field for eMammal
            }

            tasks.append((entry, full_img_path))
        
        # ...for each image
        
    # ...for each sequence
    
# ...for each deployment
    
print('Finished creating tasks to get images.')


#%% Get image sizes

# 'tasks' is currently a list of 2-tuples, with each entry as [image dictionary,path].
# 
# Go through that and copy just the image dictionaries to 'db_images', adding size
# information to each entry.  Takes a couple hours.

db_images = []
corrupt_images = []
pbar = tqdm(total=len(tasks))

if run_parallel:
    # opening each image seems too fast for this multi-threaded version to be faster than sequential code.
    num_workers = multiprocessing.cpu_count()
    pool = ThreadPool(num_workers)
    db_images = pool.starmap(_add_image, tasks)
    print('Waiting for image processes to finish...')
    pool.close()
    pool.join()
else:
    print('Finding image size sequentially')
    for entry, full_img_path in tasks:
        db_images.append(_add_image(entry, full_img_path))

db_images = [i for i in db_images if i is not None]

print('{} images could not be opened:'.format(len(corrupt_images)))
print(corrupt_images)
print('Done getting image sizes')


#%% Assemble top-level dictionaries

db_info = {
    'year': 'unknown',
    'version': version,
    'description': description,
    'contributor': contributor,
    'curator': curator,
    'date_created': str(date.today())
}

coco_formatted_json = {
    'info': db_info,
    'images': db_images
}


#%% Write out .json

print('Saving the json database to disk...')
with open(os.path.join(output_dir_path, db_filename), 'w') as f:
    json.dump(coco_formatted_json, f, indent=4, sort_keys=True)
print('...done')

print('Saving list of corrupt images...')
with open(os.path.join(output_dir_path, corrupt_images_db_filename), 'w') as f:
    json.dump(corrupt_images, f, indent=4)    
print('...done')

print('Running the script took {}.'.format(datetime.now() - start_time))





"""

md_to_coco.py

"Converts" MegaDetector output files to COCO format.  "Converts" is in quotes because
this is an opinionated transformation that requires a confidence threshold.

Does not currently handle classification information.

"""

#%% Constants and imports

import os
import json
import uuid

from tqdm import tqdm

from md_visualization import visualization_utils as vis_utils

default_confidence_threshold = 0.15


#%% Functions

def md_to_coco(md_results_file,
               coco_output_file=None,
               image_folder=None,
               confidence_threshold=default_confidence_threshold,
               validate_image_sizes=False,
               info=None,
               preserve_nonstandard_metadata=True,
               include_failed_images=True):
    """
    "Converts" MegaDetector output files to COCO format.  "Converts" is in quotes because
    this is an opinionated transformation that requires a confidence threshold.
    
    A folder of images is required if width and height information are not available 
    in the MD results file.

    Args:
        md_results_file (str): MD results .json file to convert to COCO format
        coco_output_file (str, optional): COCO .json file to write; if this is None, we'll return 
            a COCO-formatted dict, but won't write it to disk
        image_folder (str, optional): folder of images, required if 'width' and 'height' are not
            present in the MD results file (they are not required by the format)
        confidence_threshold (float, optional): boxes below this confidence threshold will not be
            included in the output data
        validate_image_sizes (bool, optional): if this is True, we'll check the image sizes 
            regardless of whether "width" and "height" are present in the MD results file.
        info (dict, optional): arbitrary metadata to include in an "info" field in the COCO-formatted
            output
        preserve_nonstandard_metadata (bool, optional): if this is True, confidence will be preserved in a 
            non-standard "conf" field in each annotation, and any random fields present in each image's data
            (e.g. EXIF metadata) will be propagated to COCO output    
        include_failed_images (boo, optional): if this is True, failed images will be propagated to COCO output 
            with a non-empty "failure" field and no other fields, otherwise failed images will be skipped.
    
    Returns:
        dict: the COCO data dict, identical to what's written to [coco_output_file] if [coco_output_file]
        is not None.
    """

    with open(md_results_file,'r') as f:    
        md_results = json.load(f)    
        
    coco_images = []
    coco_annotations = []
    
    # im = md_results['images'][0]
    for im in tqdm(md_results['images']):
        
        coco_im = {}
        coco_im['id'] = im['file']
        coco_im['file_name'] = im['file']
        
        # There is no concept of this in the COCO standard        
        if 'failure' in im and im['failure'] is not None:
            if include_failed_images:
                coco_im['failure'] = im['failure']
                coco_images.append(coco_im)
            continue
        
        # Read/validate image size
        w = None
        h = None
        
        if ('width' not in im) or ('height' not in im) or validate_image_sizes:
            if image_folder is None:
                raise ValueError('Must provide an image folder when height/width need to be read from images')
            image_file_abs = os.path.join(image_folder,im['file'])
            pil_im = vis_utils.open_image(image_file_abs)
            w = pil_im.width
            h = pil_im.height
            if validate_image_sizes:
                if 'width' in im:
                    assert im['width'] == w, 'Width mismatch for image {}'.format(im['file'])
                if 'height' in im:
                    assert im['height'] == h, 'Height mismatch for image {}'.format(im['file'])
        else:
            w = im['width']
            h = im['height']
            
        coco_im['width'] = w
        coco_im['height'] = h
        
        # Add other, non-standard fields to the output dict
        if preserve_nonstandard_metadata:
            for k in im.keys():
                if k not in ('file','detections','width','height'):
                    coco_im[k] = im[k]
                
        coco_images.append(coco_im)
        
        # detection = im['detections'][0]
        for detection in im['detections']:
            
            # Skip below-threshold detections
            if confidence_threshold is not None and detection['conf'] < confidence_threshold:
                continue
        
            # Create an annotation
            ann = {}        
            ann['id'] = str(uuid.uuid1())
            ann['image_id'] = coco_im['id']                
                        
            md_category_id = detection['category']
            coco_category_id = int(md_category_id)
            ann['category_id'] = coco_category_id
            
            # In very esoteric cases, we use the empty category (0) in MD-formatted output files
            if md_category_id != '0':
                
                assert 'bbox' in detection,\
                    'Oops: non-empty category with no bbox in {}'.format(im['file'])
            
                ann['bbox'] = detection['bbox']
                # MegaDetector: [x,y,width,height] (normalized, origin upper-left)
                # COCO: [x,y,width,height] (absolute, origin upper-left)
                ann['bbox'][0] = ann['bbox'][0] * coco_im['width']
                ann['bbox'][1] = ann['bbox'][1] * coco_im['height']
                ann['bbox'][2] = ann['bbox'][2] * coco_im['width']
                ann['bbox'][3] = ann['bbox'][3] * coco_im['height']                
            
            else:
                
                print('Warning: empty category annotation in file {}'.format(im['file']))
                      
            if preserve_nonstandard_metadata:
                ann['conf'] = detection['conf']
                
            coco_annotations.append(ann)            
        
        # ...for each detection
    
    # ...for each image

    output_dict = {}
    
    if info is not None:
        output_dict['info'] = info
    else:
        output_dict['info'] = {'description':'Converted from MD results file {}'.format(md_results_file)}
    output_dict['info']['confidence_threshold'] = confidence_threshold
    
    output_dict['images'] = coco_images
    output_dict['annotations'] = coco_annotations
    
    output_dict['categories'] = []
    
    for md_category_id in md_results['detection_categories'].keys():
        
        coco_category_id = int(md_category_id)
        coco_category = {'id':coco_category_id,
                         'name':md_results['detection_categories'][md_category_id]}
        output_dict['categories'].append(coco_category)
        
    if coco_output_file is not None:
        with open(coco_output_file,'w') as f:
            json.dump(output_dict,f,indent=1)
    
    return output_dict

# def md_to_coco(...)


#%% Interactive driver

if False:
    
    pass

    #%% Configure options
    
    md_results_file = os.path.expanduser('~/data/md-test.json')
    coco_output_file = os.path.expanduser('~/data/md-test-coco.json')
    image_folder = os.path.expanduser('~/data/md-test')
    validate_image_sizes = True
    confidence_threshold = 0.2
    validate_image_sizes=True
    info=None
    preserve_nonstandard_metadata=True
    include_failed_images=False
    
    
    #%% Programmatic execution
    
    output_dict = md_to_coco(md_results_file,
                   coco_output_file=coco_output_file,
                   image_folder=image_folder,
                   confidence_threshold=confidence_threshold,
                   validate_image_sizes=validate_image_sizes,
                   info=info,
                   preserve_nonstandard_metadata=preserve_nonstandard_metadata,
                   include_failed_images=include_failed_images)
    
    
    #%% Command-line example
        
    s = f'python md_to_coco.py {md_results_file} {coco_output_file} {confidence_threshold} '
    if image_folder is not None:
        s += f' --image_folder {image_folder}'
    if preserve_nonstandard_metadata:
        s += ' --preserve_nonstandard_metadata'
    if include_failed_images:
        s += ' --include_failed_images'

    print(s); import clipboard; clipboard.copy(s)

    
    #%% Preview the resulting file
    
    from md_visualization import visualize_db
    options = visualize_db.DbVizOptions()
    options.parallelize_rendering = True
    options.viz_size = (900, -1)
    options.num_to_visualize = 5000

    html_file,_ = visualize_db.visualize_db(coco_output_file,
                                              os.path.expanduser('~/tmp/md_to_coco_preview'),
                                              image_folder,options)    

    from md_utils import path_utils # noqa
    path_utils.open_file(html_file)
    
    
#%% Command-line driver

import sys,argparse

def main():

    parser = argparse.ArgumentParser(
        description='"Convert" MD output to COCO format, in quotes because this is an opinionated transformation that requires a confidence threshold')
    
    parser.add_argument(
        'md_results_file',
        type=str,
        help='Path to MD results file (.json)')
    
    parser.add_argument(
        'coco_output_file',
        type=str,
        help='Output filename (.json)')
    
    parser.add_argument(
        'confidence_threshold',
        type=float,
        default=default_confidence_threshold,
        help='Confidence threshold (default {})'.format(default_confidence_threshold)
        )
    
    parser.add_argument(
        '--image_folder',
        type=str,
        default=None,
        help='Image folder, only required if we will need to access image sizes'
        )
    
    parser.add_argument(
        '--preserve_nonstandard_metadata',
        action='store_true',
        help='Preserve metadata that isn\'t normally included in ' + 
             'COCO-formatted data (e.g. EXIF metadata, confidence values)'
        )

    parser.add_argument(
        '--include_failed_images',
        action='store_true',
        help='Keep a record of corrupted images in the output; may not be completely COCO-compliant'
        )
    
    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    md_to_coco(args.md_results_file,
               args.coco_output_file,
               args.image_folder,
               args.confidence_threshold,
               validate_image_sizes=False,
               info=None,
               preserve_nonstandard_metadata=args.preserve_nonstandard_metadata,
               include_failed_images=args.include_failed_images)    
    
if __name__ == '__main__':
    main()
